"""
Common global functions defined by GAP.
"""

###############################################################################
#       Copyright (C) 2009, William Stein <wstein@gmail.com>
#       Copyright (C) 2012, Volker Braun <vbraun.name@gmail.com>
#
#   Distributed under the terms of the GNU General Public License (GPL)
#   as published by the Free Software Foundation; either version 2 of
#   the License, or (at your option) any later version.
#                   https://www.gnu.org/licenses/
###############################################################################


# selected gap functions to use in tab completion
common_gap_functions = set([
  'AbelianGroup',
  'AbelianInvariants',
  'AbelianInvariantsMultiplier',
  'AbelianInvariantsOfList',
  'AbelianNumberField',
  'AbsInt',
  'AbsoluteValue',
  'Action',
  'ActionHomomorphism',
  'Add',
  'AddCoeffs',
  'AddGenerator',
  'AddRelator',
  'AddRowVector',
  'AddRule',
  'AddSet',
  'AdjointMatrix',
  'Algebra',
  'AlternatingGroup',
  'AntiSymmetricParts',
  'Append',
  'AppendTo',
  'Apply',
  'AsGroup',
  'AtlasGroup',
  'AutomorphismGroup',
  'BaseOfGroup',
  'Basis',
  'BasisVectors',
  'Bell',
  'Binomial',
  'BlockMatrix',
  'Blocks',
  'CartanMatrix',
  'CartanSubalgebra',
  'Cartesian',
  'Center',
  'CentralCharacter',
  'Centralizer',
  'CentralizerInGLnZ',
  'CentralizerModulo',
  'Centre',
  'CentreOfCharacter',
  'Character',
  'CharacterDegrees',
  'CharacterNames',
  'CharacterTable',
  'Characteristic',
  'CharacteristicPolynomial',
  'CheckFixedPoints',
  'ChevalleyBasis',
  'ChiefNormalSeriesByPcgs',
  'ChiefSeries',
  'ChineseRem',
  'Chomp',
  'ClassElementLattice',
  'ClassFunction',
  'ClassFunctionSameType',
  'ClassOrbit',
  'ClassPermutation',
  'ClassRoots',
  'ClassesSolvableGroup',
  'CoKernel',
  'Coefficients',
  'CoefficientsRing',
  'CoeffsCyc',
  'CoeffsMod',
  'CollapsedMat',
  'Collected',
  'Combinations',
  'CombinatorialCollector',
  'CommutatorFactorGroup',
  'CommutatorLength',
  'CommutatorSubgroup',
  'Compacted',
  'CompanionMat',
  'ComplexConjugate',
  'ComplexificationQuat',
  'CompositionMapping',
  'CompositionMapping2',
  'CompositionMaps',
  'Concatenation',
  'Conductor',
  'ConjugacyClass',
  'ConjugacyClassSubgroups',
  'ConjugacyClasses',
  'ConjugateGroup',
  'ConjugateSubgroup',
  'ConjugateSubgroups',
  'ConstituentsCompositionMapping',
  'ContainedMaps',
  'ContinuedFractionApproximationOfRoot',
  'ContinuedFractionExpansionOfRoot',
  'ConvertToCharacterTable',
  'ConvertToMatrixRep',
  'ConvertToRangeRep',
  'ConvertToStringRep',
  'ConvertToTableOfMarks',
  'ConvertToVectorRep',
  'ConwayPolynomial',
  'CosetTable',
  'CosetTableInWholeGroup',
  'Cycle',
  'CycleLength',
  'CycleLengths',
  'CycleStructureClass',
  'CycleStructurePerm',
  'Cycles',
  'CyclicGroup',
  'CyclotomicField',
  'CyclotomicPolynomial',
  'DefiningPolynomial',
  'Degree',
  'DegreeFFE',
  'DenominatorCyc',
  'DenominatorOfRationalFunction',
  'DenominatorRat',
  'Derivations',
  'Derivative',
  'DerivedLength',
  'DerivedSeries',
  'DerivedSeriesOfGroup',
  'DerivedSubgroup',
  'Determinant',
  'DeterminantIntMat',
  'DeterminantMat',
  'DeterminantMatDivFree',
  'DeterminantOfCharacter',
  'DiagonalMat',
  'DihedralGroup',
  'Dimension',
  'DimensionOfMatrixGroup',
  'DimensionsMat',
  'DirectProduct',
  'Discriminant',
  'Display',
  'DivisorsInt',
  'DnLattice',
  'DominantCharacter',
  'DominantWeights',
  'DoubleCoset',
  'DoubleCosetRepsAndSizes',
  'DoubleCosets',
  'DoubleHashArraySize',
  'DuplicateFreeList',
  'E',
  'Eigenspaces',
  'Eigenvalues',
  'Eigenvectors',
  'ElementOfFpGroup',
  'ElementOfFpSemigroup',
  'ElementOrdersPowerMap',
  'Elements',
  'ElementsStabChain',
  'EpimorphismFromFreeGroup',
  'EpimorphismNilpotentQuotient',
  'EpimorphismPGroup',
  'EpimorphismQuotientSystem',
  'EpimorphismSchurCover',
  'EuclideanQuotient',
  'EuclideanRemainder',
  'EulerianFunction',
  'Exponent',
  'Extension',
  'ExteriorCentre',
  'ExteriorPower',
  'Extract',
  'FactorGroup',
  'Factorial',
  'Factorization',
  'Factors',
  'FactorsInt',
  'Fibonacci',
  'Field',
  'FieldExtension',
  'FieldOfMatrixGroup',
  'Filtered',
  'First',
  'FittingSubgroup',
  'Flat',
  'ForAll',
  'ForAny',
  'FreeGroup',
  'FreeProduct',
  'FreeSemigroup',
  'FrobeniusAutomorphism',
  'FullRowSpace',
  'GF',
  'GL',
  'GQuotients',
  'GaloisCyc',
  'GaloisField',
  'GaloisGroup',
  'GaloisMat',
  'GaloisStabilizer',
  'Gcd',
  'GcdInt',
  'GcdOp',
  'GeneralLinearGroup',
  'GeneralOrthogonalGroup',
  'GeneralUnitaryGroup',
  'GeneralisedEigenspaces',
  'GeneralisedEigenvalues',
  'GeneralizedEigenspaces',
  'GeneralizedEigenvalues',
  'GeneratorsOfField',
  'GeneratorsOfGroup',
  'GeneratorsOfIdeal',
  'GroebnerBasis',
  'Group',
  'GroupHomomorphismByFunction',
  'GroupHomomorphismByImages',
  'GroupRing',
  'HermiteNormalFormIntegerMat',
  'HermiteNormalFormIntegerMatTransform',
  'Hom',
  'IdGroup',
  'Ideal',
  'IdealByGenerators',
  'Idempotents',
  'Identifier',
  'Identity',
  'Image',
  'Images',
  'Index',
  'InfoLevel',
  'InfoText',
  'InnerAutomorphism',
  'InnerAutomorphismsAutomorphismGroup',
  'Int',
  'IntFFE',
  'IntFFESymm',
  'IntHexString',
  'IntScalarProducts',
  'IntVecFFE',
  'IntersectSet',
  'Intersection',
  'InvariantBilinearForm',
  'InvariantElementaryAbelianSeries',
  'InvariantLattice',
  'InvariantQuadraticForm',
  'InvariantSesquilinearForm',
  'Inverse',
  'InverseMap',
  'Irr',
  'IrrBaumClausen',
  'IrrConlon',
  'IrrDixonSchneider',
  'IrreducibleModules',
  'IrreducibleRepresentations',
  'IrreducibleRepresentationsDixon',
  'IsAbelian',
  'IsAbelianNumberField',
  'IsAbelianNumberFieldPolynomialRing',
  'IsAdditiveElement',
  'IsAdditiveElementWithInverse',
  'IsAdditiveElementWithZero',
  'IsAdditiveGroup',
  'IsAdditiveGroupGeneralMapping',
  'IsAdditiveGroupHomomorphism',
  'IsAdditivelyCommutative',
  'IsAdditivelyCommutativeElement',
  'IsAlgebra',
  'IsAlgebraGeneralMapping',
  'IsAlgebraHomomorphism',
  'IsAlgebraModule',
  'IsAlgebraWithOne',
  'IsAlgebraWithOneHomomorphism',
  'IsAlgebraicElement',
  'IsAlgebraicExtension',
  'IsAlternatingGroup',
  'IsAnticommutative',
  'IsAntisymmetricBinaryRelation',
  'IsAssocWord',
  'IsAssocWordWithInverse',
  'IsAssocWordWithOne',
  'IsAssociated',
  'IsAssociative',
  'IsAutomorphismGroup',
  'IsBasis',
  'IsBijective',
  'IsBinaryRelation',
  'IsBlockMatrixRep',
  'IsBool',
  'IsBoundGlobal',
  'IsBrauerTable',
  'IsBravaisGroup',
  'IsBuiltFromGroup',
  'IsBuiltFromSemigroup',
  'IsCanonicalBasis',
  'IsCanonicalBasisFullMatrixModule',
  'IsCanonicalBasisFullRowModule',
  'IsCanonicalNiceMonomorphism',
  'IsCentral',
  'IsCentralFactor',
  'IsChar',
  'IsCharacter',
  'IsCharacterTable',
  'IsCharacterTableInProgress',
  'IsCharacteristicSubgroup',
  'IsClosedStream',
  'IsCochain',
  'IsCochainCollection',
  'IsCommutative',
  'IsComponentObjectRep',
  'IsCompositionMappingRep',
  'IsConfluent',
  'IsConjugate',
  'IsCopyable',
  'IsCyc',
  'IsCyclic',
  'IsCyclotomic',
  'IsCyclotomicField',
  'IsCyclotomicMatrixGroup',
  'IsDenseList',
  'IsDiagonalMat',
  'IsDictionary',
  'IsDigitChar',
  'IsDivisionRing',
  'IsDomain',
  'IsDoneIterator',
  'IsDoubleCoset',
  'IsDuplicateFree',
  'IsDuplicateFreeList',
  'IsElementaryAbelian',
  'IsEmpty',
  'IsEmptyString',
  'IsEuclideanRing',
  'IsFFE',
  'IsField',
  'IsFinite',
  'IsFiniteDimensional',
  'IsFinitelyGeneratedGroup',
  'IsFixedStabilizer',
  'IsFpGroup',
  'IsFpMonoid',
  'IsFpSemigroup',
  'IsFreeGroup',
  'IsFreeLeftModule',
  'IsFullHomModule',
  'IsFullMatrixModule',
  'IsFullRowModule',
  'IsFunction',
  'IsGL',
  'IsGaussInt',
  'IsGaussRat',
  'IsGaussianIntegers',
  'IsGaussianRationals',
  'IsGaussianSpace',
  'IsGeneralLinearGroup',
  'IsGroup',
  'IsGroupHomomorphism',
  'IsGroupOfAutomorphisms',
  'IsGroupRing',
  'IsHasseDiagram',
  'IsHomogeneousList',
  'IsIdempotent',
  'IsInfinity',
  'IsInjective',
  'IsInnerAutomorphism',
  'IsInt',
  'IsIntegerMatrixGroup',
  'IsIntegers',
  'IsIntegralBasis',
  'IsIntegralCyclotomic',
  'IsIntegralRing',
  'IsIrreducible',
  'IsIrreducibleCharacter',
  'IsIrreducibleRingElement',
  'IsIterator',
  'IsJacobianRing',
  'IsLaurentPolynomial',
  'IsLaurentPolynomialDefaultRep',
  'IsLexicographicallyLess',
  'IsLieAbelian',
  'IsLieAlgebra',
  'IsLieMatrix',
  'IsLieObject',
  'IsLieObjectCollection',
  'IsLieSolvable',
  'IsLinearMapping',
  'IsLinearMappingsModule',
  'IsList',
  'IsMapping',
  'IsMatchingSublist',
  'IsMatrix',
  'IsMatrixGroup',
  'IsMatrixModule',
  'IsMatrixSpace',
  'IsMonomial',
  'IsMonomialGroup',
  'IsMonomialMatrix',
  'IsMonomialOrdering',
  'IsMultiplicativeZero',
  'IsMutable',
  'IsMutableBasis',
  'IsNilpotent',
  'IsNilpotentElement',
  'IsNilpotentGroup',
  'IsNormal',
  'IsNormalBasis',
  'IsNotIdenticalObj',
  'IsNumberField',
  'IsObject',
  'IsOddInt',
  'IsOne',
  'IsOrdering',
  'IsOrdinaryMatrix',
  'IsOrdinaryTable',
  'IsPGroup',
  'IsPSolvable',
  'IsPcGroup',
  'IsPcgs',
  'IsPerfect',
  'IsPerfectGroup',
  'IsPerm',
  'IsPermGroup',
  'IsPolycyclicGroup',
  'IsPolynomial',
  'IsPolynomialRing',
  'IsPosInt',
  'IsPosRat',
  'IsPositiveIntegers',
  'IsPrime',
  'IsPrimeField',
  'IsPrimeInt',
  'IsPrimePowerInt',
  'IsPrimitive',
  'IsPrimitiveCharacter',
  'IsPrimitivePolynomial',
  'IsProbablyPrimeInt',
  'IsPurePadicNumber',
  'IsQuaternion',
  'IsQuickPositionList',
  'IsQuotientSemigroup',
  'IsRandomSource',
  'IsRange',
  'IsRat',
  'IsRationalFunction',
  'IsRationalMatrixGroup',
  'IsRationals',
  'IsRecord',
  'IsReduced',
  'IsReductionOrdering',
  'IsReflexiveBinaryRelation',
  'IsRegular',
  'IsRegularSemigroup',
  'IsRegularSemigroupElement',
  'IsRing',
  'IsRingElement',
  'IsRingGeneralMapping',
  'IsRingWithOne',
  'IsRingWithOneGeneralMapping',
  'IsRingWithOneHomomorphism',
  'IsRowModule',
  'IsRowSpace',
  'IsRowVector',
  'IsSL',
  'IsSSortedList',
  'IsScalar',
  'IsSet',
  'IsSimple',
  'IsSimpleAlgebra',
  'IsSimpleGroup',
  'IsSimpleSemigroup',
  'IsSingleValued',
  'IsSolvable',
  'IsSolvableGroup',
  'IsSortedList',
  'IsSpecialLinearGroup',
  'IsSporadicSimple',
  'IsStandardIterator',
  'IsString',
  'IsStringRep',
  'IsSubgroup',
  'IsSubgroupFpGroup',
  'IsSubgroupOfWholeGroupByQuotientRep',
  'IsSubgroupSL',
  'IsSubset',
  'IsSubsetSet',
  'IsSubspace',
  'IsSupersolvable',
  'IsSupersolvableGroup',
  'IsSurjective',
  'IsSymmetricGroup',
  'IsTable',
  'IsTotal',
  'IsTotalOrdering',
  'IsTransformation',
  'IsTransitive',
  'IsTransitiveBinaryRelation',
  'IsTrivial',
  'IsUniqueFactorizationRing',
  'IsUnit',
  'IsUnivariatePolynomial',
  'IsUnivariatePolynomialRing',
  'IsUnivariateRationalFunction',
  'IsUpperAlphaChar',
  'IsUpperTriangularMat',
  'IsValidIdentifier',
  'IsVector',
  'IsVectorSpace',
  'IsVirtualCharacter',
  'IsWeylGroup',
  'IsWord',
  'IsZero',
  'IsZeroGroup',
  'IsZeroSimpleSemigroup',
  'IsZeroSquaredRing',
  'IsZmodnZObj',
  'IsZmodnZObjNonprime',
  'IsZmodpZObj',
  'IsZmodpZObjLarge',
  'IsZmodpZObjSmall',
  'IsomorphicSubgroups',
  'IsomorphismFpAlgebra',
  'IsomorphismFpGroup',
  'IsomorphismFpGroupByGenerators',
  'IsomorphismFpGroupByPcgs',
  'IsomorphismFpSemigroup',
  'IsomorphismGroups',
  'IsomorphismMatrixAlgebra',
  'IsomorphismPcGroup',
  'IsomorphismPermGroup',
  'IsomorphismPermGroupImfGroup',
  'IsomorphismReesMatrixSemigroup',
  'IsomorphismRefinedPcGroup',
  'IsomorphismSimplifiedFpGroup',
  'IsomorphismSpecialPcGroup',
  'IsomorphismTransformationSemigroup',
  'IsomorphismTypeInfoFiniteSimpleGroup',
  'Iterated',
  'Iterator',
  'IteratorByBasis',
  'IteratorByFunctions',
  'IteratorList',
  'IteratorSorted',
  'Jacobi',
  'JenningsLieAlgebra',
  'JenningsSeries',
  'JordanDecomposition',
  'Kernel',
  'KernelOfAdditiveGeneralMapping',
  'KernelOfCharacter',
  'KernelOfMultiplicativeGeneralMapping',
  'KernelOfTransformation',
  'KillingMatrix',
  'KnuthBendixRewritingSystem',
  'KroneckerProduct',
  'KuKGenerators',
  'LLL',
  'LLLReducedBasis',
  'LLLReducedGramMat',
  'Lambda',
  'LargestElementGroup',
  'LargestElementStabChain',
  'LargestMovedPoint',
  'LastSystemError',
  'LatticeByCyclicExtension',
  'LatticeSubgroups',
  'Lcm',
  'LcmInt',
  'LcmOp',
  'LeadingCoefficient',
  'LeadingCoefficientOfPolynomial',
  'LeadingExponentOfPcElement',
  'LeadingMonomial',
  'LeadingMonomialOfPolynomial',
  'LeadingTermOfPolynomial',
  'Legendre',
  'Length',
  'LenstraBase',
  'LessThanFunction',
  'LessThanOrEqualFunction',
  'LetterRepAssocWord',
  'LevelsOfGenerators',
  'LeviMalcevDecomposition',
  'LexicographicOrdering',
  'LieAlgebra',
  'LieAlgebraByStructureConstants',
  'LieBracket',
  'LieCenter',
  'LieCentralizer',
  'LieCentre',
  'LieCoboundaryOperator',
  'LieDerivedSeries',
  'LieDerivedSubalgebra',
  'LieLowerCentralSeries',
  'LieNilRadical',
  'LieNormalizer',
  'LieObject',
  'LieSolvableRadical',
  'LieUpperCentralSeries',
  'LiftedInducedPcgs',
  'LiftedPcElement',
  'LinearAction',
  'LinearActionLayer',
  'LinearCharacters',
  'LinearCombination',
  'LinearCombinationPcgs',
  'LinearIndependentColumns',
  'LinearOperation',
  'LinearOperationLayer',
  'LinesOfStraightLineProgram',
  'List',
  'ListN',
  'ListPerm',
  'ListStabChain',
  'ListWithIdenticalEntries',
  'ListX',
  'LoadDynamicModule',
  'LoadPackage',
  'Log',
  'LogFFE',
  'LogInt',
  'LogMod',
  'LogModShanks',
  'LogTo',
  'LongestWeylWordPerm',
  'LookupDictionary',
  'LowIndexSubgroupsFpGroup',
  'LowIndexSubgroupsFpGroupIterator',
  'LowerCentralSeries',
  'LowerCentralSeriesOfGroup',
  'LowercaseString',
  'Lucas',
  'MakeConfluent',
  'MakeImmutable',
  'MakeReadOnlyGlobal',
  'MakeReadWriteGlobal',
  'MappedWord',
  'MappingByFunction',
  'MappingPermListList',
  'MatAlgebra',
  'MatClassMultCoeffsCharTable',
  'MatLieAlgebra',
  'MatScalarProducts',
  'MathieuGroup',
  'MatrixAlgebra',
  'MatrixAutomorphisms',
  'MatrixByBlockMatrix',
  'MatrixLieAlgebra',
  'MatrixOfAction',
  'MaximalAbelianQuotient',
  'MaximalBlocks',
  'MaximalNormalSubgroups',
  'MaximalSubgroupClassReps',
  'MaximalSubgroups',
  'MaximalSubgroupsLattice',
  'Maximum',
  'MaximumList',
  'MeetEquivalenceRelations',
  'MeetMaps',
  'MinimalElementCosetStabChain',
  'MinimalGeneratingSet',
  'MinimalNonmonomialGroup',
  'MinimalNormalSubgroups',
  'MinimalPolynomial',
  'MinimalStabChain',
  'MinimalSupergroupsLattice',
  'MinimizedBombieriNorm',
  'Minimum',
  'MinimumList',
  'MinusCharacter',
  'ModuleByRestriction',
  'ModuleOfExtension',
  'ModuloPcgs',
  'MoebiusMu',
  'MolienSeries',
  'MolienSeriesInfo',
  'MolienSeriesWithGivenDenominator',
  'Monoid',
  'MonoidByGenerators',
  'MonoidByMultiplicationTable',
  'MonoidOfRewritingSystem',
  'MonomialComparisonFunction',
  'MonomialExtGrlexLess',
  'MonomialExtrepComparisonFun',
  'MonomialGrevlexOrdering',
  'MonomialGrlexOrdering',
  'MonomialLexOrdering',
  'MostFrequentGeneratorFpGroup',
  'MovedPoints',
  'MultRowVector',
  'MultiplicationTable',
  'MultiplicativeNeutralElement',
  'MultiplicativeZero',
  'MultiplicativeZeroOp',
  'NF',
  'NK',
  'NameFunction',
  'NaturalCharacter',
  'NaturalHomomorphismByGenerators',
  'NaturalHomomorphismByIdeal',
  'NaturalHomomorphismByNormalSubgroup',
  'NaturalHomomorphismBySubAlgebraModule',
  'NaturalHomomorphismBySubspace',
  'NearAdditiveGroup',
  'NearAdditiveGroupByGenerators',
  'NegativeRootVectors',
  'NegativeRoots',
  'NextIterator',
  'NextPrimeInt',
  'NiceBasis',
  'NiceFreeLeftModule',
  'NiceFreeLeftModuleInfo',
  'NiceMonomorphism',
  'NiceMonomorphismAutomGroup',
  'NiceObject',
  'NiceVector',
  'NilpotencyClassOfGroup',
  'NonabelianExteriorSquare',
  'Norm',
  'NormalBase',
  'NormalClosure',
  'NormalFormIntMat',
  'NormalIntersection',
  'NormalSeriesByPcgs',
  'NormalSubgroups',
  'NormalizeWhitespace',
  'NormalizedWhitespace',
  'Normalizer',
  'NormalizerInGLnZ',
  'NormalizerInGLnZBravaisGroup',
  'NormedRowVector',
  'NrArrangements',
  'NrBasisVectors',
  'NrCombinations',
  'NrConjugacyClasses',
  'NrConjugacyClassesGL',
  'NrConjugacyClassesGU',
  'NrConjugacyClassesPGL',
  'NrConjugacyClassesPGU',
  'NrConjugacyClassesPSL',
  'NrConjugacyClassesPSU',
  'NrConjugacyClassesSL',
  'NrConjugacyClassesSLIsogeneous',
  'NrConjugacyClassesSU',
  'NrConjugacyClassesSUIsogeneous',
  'NrDerangements',
  'NrInputsOfStraightLineProgram',
  'NrMovedPoints',
  'NrOrderedPartitions',
  'NrPartitionTuples',
  'NrPartitions',
  'NrPartitionsSet',
  'NrPermutationsList',
  'NrPolyhedralSubgroups',
  'NrPrimitiveGroups',
  'NrRestrictedPartitions',
  'NrTransitiveGroups',
  'NrTuples',
  'NrUnorderedTuples',
  'NullAlgebra',
  'NullMat',
  'NullspaceIntMat',
  'NullspaceMat',
  'NullspaceMatDestructive',
  'NullspaceModQ',
  'Number',
  'NumberArgumentsFunction',
  'NumberFFVector',
  'NumberPerfectGroups',
  'NumberPerfectLibraryGroups',
  'NumberSmallGroups',
  'NumberSyllables',
  'NumeratorOfModuloPcgs',
  'NumeratorOfRationalFunction',
  'NumeratorRat',
  'Objectify',
  'ObjectifyWithAttributes',
  'OctaveAlgebra',
  'OldGeneratorsOfPresentation',
  'Omega',
  'OnBreak',
  'OnBreakMessage',
  'OnIndeterminates',
  'OnLeftInverse',
  'OnLines',
  'OnPairs',
  'OnPoints',
  'OnRight',
  'OnSets',
  'OnSetsDisjointSets',
  'OnSetsSets',
  'OnSetsTuples',
  'OnSubspacesByCanonicalBasis',
  'OnTuples',
  'OnTuplesSets',
  'OnTuplesTuples',
  'One',
  'OneAttr',
  'OneCoboundaries',
  'OneCocycles',
  'OneFactorBound',
  'OneImmutable',
  'OneMutable',
  'OneOfPcgs',
  'OneOp',
  'OneSM',
  'OneSameMutability',
  'OperationAlgebraHomomorphism',
  'Orbit',
  'OrbitFusions',
  'OrbitLength',
  'OrbitLengths',
  'OrbitLengthsDomain',
  'OrbitPerms',
  'OrbitPowerMaps',
  'OrbitStabChain',
  'OrbitStabilizer',
  'OrbitStabilizerAlgorithm',
  'Orbits',
  'OrbitsDomain',
  'OrbitsPerms',
  'Order',
  'OrderMod',
  'OrderingOnGenerators',
  'Ordinal',
  'OrdinaryCharacterTable',
  'OrthogonalComponents',
  'OrthogonalEmbeddings',
  'OrthogonalEmbeddingsSpecialDimension',
  'PCentralLieAlgebra',
  'PCentralNormalSeriesByPcgsPGroup',
  'PCentralSeries',
  'PClassPGroup',
  'PCore',
  'PGL',
  'PGU',
  'POW',
  'PQuotient',
  'PROD',
  'PSL',
  'PSP',
  'PSU',
  'PSp',
  'PadicCoefficients',
  'PadicNumber',
  'PadicValuation',
  'Parametrized',
  'Parent',
  'ParentPcgs',
  'PartialFactorization',
  'PartialOrderByOrderingFunction',
  'PartialOrderOfHasseDiagram',
  'Partition',
  'PartitionTuples',
  'Partitions',
  'PartitionsGreatestEQ',
  'PartitionsGreatestLE',
  'PartitionsSet',
  'PcGroupCode',
  'PcGroupCodeRec',
  'PcGroupFpGroup',
  'PcGroupWithPcgs',
  'PcSeries',
  'Pcgs',
  'PcgsCentralSeries',
  'PcgsChiefSeries',
  'PcgsElementaryAbelianSeries',
  'PcgsPCentralSeriesPGroup',
  'Pcgs_OrbitStabilizer',
  'PerfectGroup',
  'PerfectIdentification',
  'PerfectResiduum',
  'Perform',
  'PermBounds',
  'PermCharInfo',
  'PermCharInfoRelative',
  'PermChars',
  'PermComb',
  'PermLeftQuoTransformation',
  'PermList',
  'PermListList',
  'Permanent',
  'Permutation',
  'PermutationCharacter',
  'PermutationCycle',
  'PermutationCycleOp',
  'PermutationGModule',
  'PermutationMat',
  'PermutationsList',
  'Permuted',
  'Phi',
  'PolynomialByExtRep',
  'PolynomialCoefficientsOfPolynomial',
  'PolynomialDivisionAlgorithm',
  'PolynomialModP',
  'PolynomialReducedRemainder',
  'PolynomialReduction',
  'PolynomialRing',
  'PopOptions',
  'Position',
  'PositionBound',
  'PositionCanonical',
  'PositionNonZero',
  'PositionNot',
  'PositionNthOccurrence',
  'PositionProperty',
  'PositionSet',
  'PositionSorted',
  'PositionStream',
  'PositionSublist',
  'PositionWord',
  'PositionsOp',
  'PositiveRoots',
  'PossibleClassFusions',
  'PossiblePowerMaps',
  'PowerMap',
  'PowerMapOp',
  'PowerModCoeffs',
  'PowerModInt',
  'PowerPartition',
  'PreImage',
  'PreImageElm',
  'PreImages',
  'PreImagesElm',
  'PreImagesRange',
  'PreImagesRepresentative',
  'PreImagesSet',
  'PrefrattiniSubgroup',
  'PreimagesOfTransformation',
  'PresentationFpGroup',
  'PresentationNormalClosure',
  'PresentationNormalClosureRrs',
  'PresentationSubgroup',
  'PresentationSubgroupMtc',
  'PresentationSubgroupRrs',
  'PresentationViaCosetTable',
  'PrevPrimeInt',
  'PrimaryGeneratorWords',
  'PrimeBlocks',
  'PrimeBlocksOp',
  'PrimeField',
  'PrimePGroup',
  'PrimePowersInt',
  'PrimeResidues',
  'PrimitiveElement',
  'PrimitiveGroup',
  'PrimitiveIdentification',
  'PrimitivePolynomial',
  'PrimitiveRoot',
  'PrimitiveRootMod',
  'Print',
  'PrintAmbiguity',
  'PrintArray',
  'PrintCharacterTable',
  'PrintFactorsInt',
  'PrintFormattingStatus',
  'PrintHashWithNames',
  'PrintObj',
  'PrintTo',
  'Process',
  'Product',
  'ProductCoeffs',
  'ProductSpace',
  'ProductX',
  'ProjectedInducedPcgs',
  'ProjectedPcElement',
  'Projection',
  'ProjectionMap',
  'ProjectiveActionHomomorphismMatrixGroup',
  'ProjectiveActionOnFullSpace',
  'ProjectiveGeneralLinearGroup',
  'ProjectiveGeneralUnitaryGroup',
  'ProjectiveOrder',
  'ProjectiveSpecialLinearGroup',
  'ProjectiveSpecialUnitaryGroup',
  'ProjectiveSymplecticGroup',
  'PseudoRandom',
  'PthPowerImage',
  'PthPowerImages',
  'PushOptions',
  'QUO',
  'Quadratic',
  'QuaternionAlgebra',
  'QuoInt',
  'QuotRemLaurpols',
  'Quotient',
  'QuotientMod',
  'QuotientPolynomialsExtRep',
  'QuotientRemainder',
  'READ',
  'RadicalGroup',
  'RadicalOfAlgebra',
  'Random',
  'RandomBinaryRelationOnPoints',
  'RandomHashKey',
  'RandomInvertibleMat',
  'RandomIsomorphismTest',
  'RandomList',
  'RandomMat',
  'RandomPrimitivePolynomial',
  'RandomSource',
  'RandomTransformation',
  'RandomUnimodularMat',
  'Range',
  'Rank',
  'RankAction',
  'RankFilter',
  'RankMat',
  'RankOfTransformation',
  'RankPGroup',
  'Rat',
  'RationalClass',
  'RationalClasses',
  'RationalizedMat',
  'Read',
  'ReadAll',
  'ReadAllLine',
  'ReadAsFunction',
  'ReadByte',
  'ReadLine',
  'ReadPackage',
  'ReadPkg',
  'ReadTest',
  'RealClasses',
  'RealPart',
  'RealizableBrauerCharacters',
  'RecFields',
  'RecNames',
  'RedispatchOnCondition',
  'ReduceCoeffs',
  'ReduceCoeffsMod',
  'ReduceRules',
  'ReduceStabChain',
  'Reduced',
  'ReducedAdditiveInverse',
  'ReducedCharacters',
  'ReducedClassFunctions',
  'ReducedComm',
  'ReducedConfluentRewritingSystem',
  'ReducedConjugate',
  'ReducedDifference',
  'ReducedForm',
  'ReducedGroebnerBasis',
  'ReducedInverse',
  'ReducedLeftQuotient',
  'ReducedOne',
  'ReducedPcElement',
  'ReducedPower',
  'ReducedProduct',
  'ReducedQuotient',
  'ReducedScalarProduct',
  'ReducedSum',
  'ReducedZero',
  'Ree',
  'ReeGroup',
  'ReesCongruenceOfSemigroupIdeal',
  'ReesMatrixSemigroup',
  'ReesMatrixSemigroupElement',
  'ReesZeroMatrixSemigroup',
  'ReesZeroMatrixSemigroupElement',
  'RefinedPcGroup',
  'RegularActionHomomorphism',
  'RegularModule',
  'RelationsOfFpSemigroup',
  'RelativeBasis',
  'RelativeOrders',
  'RelatorsOfFpGroup',
  'RemInt',
  'Remove',
  'RemoveCharacters',
  'RemoveFile',
  'RemoveOuterCoeffs',
  'RemoveRelator',
  'RemoveSet',
  'RemoveStabChain',
  'ReplacedString',
  'Representative',
  'RepresentativeAction',
  'RepresentativeLinearOperation',
  'RepresentativeSmallest',
  'RepresentativesContainedRightCosets',
  'RepresentativesFusions',
  'RepresentativesMinimalBlocks',
  'RepresentativesPerfectSubgroups',
  'RepresentativesPowerMaps',
  'RepresentativesSimpleSubgroups',
  'Reread',
  'RereadPackage',
  'Reset',
  'RestrictOutputsOfSLP',
  'Restricted',
  'RestrictedClassFunction',
  'RestrictedClassFunctions',
  'RestrictedMapping',
  'RestrictedPartitions',
  'RestrictedPerm',
  'RestrictedTransformation',
  'ResultOfStraightLineProgram',
  'Resultant',
  'Reversed',
  'RewriteWord',
  'RightCoset',
  'RightCosets',
  'RightDerivations',
  'Ring',
  'RingWithOne',
  'Root',
  'RootInt',
  'RootMod',
  'RootOfDefiningPolynomial',
  'RootSystem',
  'RootsMod',
  'RoundCyc',
  'Rules',
  'SL',
  'SO',
  'SP',
  'SQ',
  'SSortedList',
  'SU',
  'SameBlock',
  'SaveWorkspace',
  'ScalarProduct',
  'SchurCover',
  'SemiSimpleType',
  'SemidirectProduct',
  'Semigroup',
  'Set',
  'SetAssertionLevel',
  'SetCommutator',
  'SetConjugate',
  'SetCrystGroupDefaultAction',
  'SetEntrySCTable',
  'SetFilterObj',
  'SetHashEntry',
  'SetHashEntryAtLastIndex',
  'SetHelpViewer',
  'SetIndeterminateName',
  'SetInfoLevel',
  'SetName',
  'SetParent',
  'SetPower',
  'ShallowCopy',
  'ShiftedCoeffs',
  'ShiftedPadicNumber',
  'ShortLexOrdering',
  'ShortestVectors',
  'Sigma',
  'SignInt',
  'SignPartition',
  'SignPerm',
  'SimpleLieAlgebra',
  'SimpleSystem',
  'SimplifiedFpGroup',
  'SimplifyPresentation',
  'SimultaneousEigenvalues',
  'SingleCollector',
  'Size',
  'SizeConsiderFunction',
  'SizeNumbersPerfectGroups',
  'SizeOfFieldOfDefinition',
  'SizeScreen',
  'SizeStabChain',
  'SizesCentralizers',
  'SizesConjugacyClasses',
  'SizesPerfectGroups',
  'SmallGeneratingSet',
  'SmallGroup',
  'SmallerDegreePermutationRepresentation',
  'SmallestGeneratorPerm',
  'SmallestMovedPoint',
  'SmallestRootInt',
  'SmithNormalFormIntegerMat',
  'Socle',
  'SocleTypePrimitiveGroup',
  'SolutionIntMat',
  'SolutionMat',
  'SolutionMatDestructive',
  'SolutionNullspaceIntMat',
  'Sort',
  'SortParallel',
  'SortedCharacterTable',
  'SortedCharacters',
  'SortedList',
  'SortedSparseActionHomomorphism',
  'SortingPerm',
  'Sp',
  'SparseActionHomomorphism',
  'SparseCartanMatrix',
  'SparseHashTable',
  'SparseIntKey',
  'SpecialLinearGroup',
  'SpecialOrthogonalGroup',
  'SpecialPcgs',
  'SpecialUnitaryGroup',
  'SplitCharacters',
  'SplitExtension',
  'SplitString',
  'SplittingField',
  'Sqrt',
  'SquareRoots',
  'StabChain',
  'StabChainBaseStrongGenerators',
  'StabChainImmutable',
  'StabChainMutable',
  'StabChainOp',
  'StabChainOptions',
  'Stabilizer',
  'StabilizerOfExternalSet',
  'StabilizerPcgs',
  'StandardAssociate',
  'StandardizeTable',
  'StarCyc',
  'Stirling1',
  'Stirling2',
  'StratMeetPartition',
  'StretchImportantSLPElement',
  'String',
  'StringDate',
  'StringOfResultOfStraightLineProgram',
  'StringPP',
  'StringTime',
  'StructuralCopy',
  'StructureConstantsTable',
  'StructureDescription',
  'SubAlgebraModule',
  'Subalgebra',
  'SubdirectProduct',
  'SubdirectProducts',
  'Subfield',
  'Subfields',
  'Subgroup',
  'SubgroupByPcgs',
  'SubgroupByProperty',
  'SubgroupOfWholeGroupByCosetTable',
  'SubgroupOfWholeGroupByQuotientSubgroup',
  'SubgroupProperty',
  'SubgroupShell',
  'SubgroupsSolvableGroup',
  'Submodule',
  'Submonoid',
  'SubnearAdditiveGroup',
  'SubnormalSeries',
  'Subring',
  'SubringWithOne',
  'Subsemigroup',
  'Subspace',
  'Subspaces',
  'SubstitutedWord',
  'SubtractSet',
  'Subword',
  'Successors',
  'Sum',
  'SumFactorizationFunctionPcgs',
  'SumIntersectionMat',
  'SumX',
  'SupersolvableResiduum',
  'SurjectiveActionHomomorphismAttr',
  'SuzukiGroup',
  'SylowComplement',
  'SylowSubgroup',
  'SylowSystem',
  'SymmetricClosureBinaryRelation',
  'SymmetricGroup',
  'SymmetricParentGroup',
  'SymmetricParts',
  'SymmetricPower',
  'SymmetricPowerOfAlgebraModule',
  'Symmetrizations',
  'SymplecticComponents',
  'SymplecticGroup',
  'TableAutomorphisms',
  'TableOfMarks',
  'TableOfMarksByLattice',
  'TableOfMarksCyclic',
  'TableOfMarksDihedral',
  'TableOfMarksFrobenius',
  'Tau',
  'TensorProduct',
  'TensorProductGModule',
  'TensorProductOfAlgebraModules',
  'Tensored',
  'TietzeWordAbstractWord',
  'Trace',
  'TraceImmediateMethods',
  'TraceMat',
  'TraceMethods',
  'TracePolynomial',
  'TracedCosetFpGroup',
  'TransferDiagram',
  'Transformation',
  'TransformingPermutations',
  'TransformingPermutationsCharacterTables',
  'TransitiveClosureBinaryRelation',
  'TransitiveIdentification',
  'Transitivity',
  'TranslatorSubalgebra',
  'TransposedMat',
  'TransposedMatAttr',
  'TransposedMatDestructive',
  'TransposedMatImmutable',
  'TransposedMatMutable',
  'TransposedMatOp',
  'TransposedMatrixGroup',
  'TriangulizeIntegerMat',
  'TriangulizeMat',
  'TriangulizedIntegerMat',
  'TriangulizedIntegerMatTransform',
  'TriangulizedNullspaceMat',
  'TriangulizedNullspaceMatDestructive',
  'TrivialCharacter',
  'TrivialGroup',
  'TrivialIterator',
  'TrivialSubalgebra',
  'TrivialSubgroup',
  'TrivialSubmagmaWithOne',
  'TrivialSubmodule',
  'TrivialSubmonoid',
  'TrivialSubspace',
  'Tuple',
  'Tuples',
  'TypeObj',
  'UnbindElmWPObj',
  'UnbindGlobal',
  'UnderlyingCharacterTable',
  'UnderlyingCharacteristic',
  'UnderlyingElement',
  'UnderlyingElementOfReesMatrixSemigroupElement',
  'UnderlyingElementOfReesZeroMatrixSemigroupElement',
  'UnderlyingExternalSet',
  'UnderlyingGeneralMapping',
  'UnderlyingGroup',
  'UnderlyingLeftModule',
  'UnderlyingLieAlgebra',
  'UnderlyingRelation',
  'Union',
  'Union2',
  'Unique',
  'UniteSet',
  'Units',
  'UnivariatePolynomial',
  'UnivariatePolynomialByCoefficients',
  'UnivariatePolynomialRing',
  'UnivariateRationalFunctionByCoefficients',
  'UnivariatenessTestRationalFunction',
  'UniversalEnvelopingAlgebra',
  'Unknown',
  'UnorderedTuples',
  'UnprofileFunctions',
  'UnprofileMethods',
  'UntraceMethods',
  'UpdateMap',
  'UpperCentralSeries',
  'UpperCentralSeriesOfGroup',
  'UpperSubdiagonal',
  'UseBasis',
  'UseFactorRelation',
  'UseIsomorphismRelation',
  'UseSubsetRelation',
  'Valuation',
  'Value',
  'ValueCochain',
  'ValueGlobal',
  'ValueMolienSeries',
  'ValueOption',
  'ValuePol',
  'ValuesOfClassFunction',
  'VectorSpace',
  'VectorSpaceByPcgsOfElementaryAbelianGroup',
  'View',
  'VirtualCharacter',
  'WeakPointerObj',
  'WedgeGModule',
  'WeekDay',
  'WeightLexOrdering',
  'WeightOfGenerators',
  'WeightVecFFE',
  'WeylGroup',
  'WeylOrbitIterator',
  'Where',
  'WreathProduct',
  'WreathProductImprimitiveAction',
  'WreathProductOrdering',
  'WreathProductProductAction',
  'WriteAll',
  'WriteByte',
  'WriteLine',
  'ZClassRepsQClass',
  'Zero',
  'ZeroAttr',
  'ZeroCoefficient',
  'ZeroCoefficientRatFun',
  'ZeroMapping',
  'ZeroMutable',
  'ZeroOp',
  'ZeroSM',
  'ZeroSameMutability',
  'GASMAN_STATS',
  'GASMAN',
 ])



# cython: old_style_globals=True
"""
Interpreter reset
"""
import sys

# Exclude these from the reset command.
# DATA, base64 -- needed by the notebook
# Add exit and quit to EXCLUDE to resolve Issue #22529 and Issue #16704
EXCLUDE = set(['sage_mode', '__DIR__', 'DIR', 'DATA', 'base64', 'exit', 'quit'])


def reset(vars=None, attached=False):
    """
    Delete all user-defined variables, reset all global variables
    back to their default states, and reset all interfaces to other
    computer algebra systems.

    If vars is specified, just restore the value of vars and leave
    all other variables alone (i.e., call restore).

    Note that the variables in the set :obj:`sage.misc.reset.EXCLUDE` are
    excluded from being reset.

    INPUT:

    - ``vars`` -- list or space or comma separated string (default:
      ``None``); variables to restore

    - ``attached`` -- boolean (default: ``False``); if ``vars`` is not ``None``,
      whether to detach all attached files

    EXAMPLES::

        sage: x = 5
        sage: reset()
        sage: x                                                                         # needs sage.symbolic
        x

        sage: fn = tmp_filename(ext='.py')
        sage: sage.misc.reset.EXCLUDE.add('fn')
        sage: with open(fn, 'w') as f:
        ....:     _ = f.write('a = 111')
        sage: attach(fn)
        sage: af = attached_files(); len(af)
        1
        sage: af == [fn]
        True
        sage: reset()
        sage: af = attached_files(); len(af)
        1
        sage: reset(attached=True)
        sage: af = attached_files(); len(af)
        0
        sage: sage.misc.reset.EXCLUDE.remove('fn')

    TESTS:

    Confirm that assumptions do not survive a reset (:issue:`10855`)::

        sage: # needs sage.symbolic
        sage: assume(x > 3)
        sage: assumptions()
        [x > 3]
        sage: bool(x > 3)
        True
        sage: reset()
        sage: assumptions()
        []
        sage: bool(x > 3)
        False
    """
    from sage.symbolic.assumptions import forget
    if vars is not None:
        restore(vars)
        return
    G = globals()  # this is the reason the code must be in Cython.
    T = type(sys)
    for k in list(G):
        if k[0] != '_' and not isinstance(k, T) and k not in EXCLUDE:
            try:
                del G[k]
            except KeyError:
                pass
    restore()
    forget()
    reset_interfaces()
    if attached:
        import sage.repl.attach
        sage.repl.attach.reset()


def restore(vars=None):
    """
    Restore predefined global variables to their default values.

    INPUT:

    - ``vars`` -- string or list (default: ``None``); if not ``None``, restores
      just the given variables to the default value

    EXAMPLES::

        sage: x = 10; y = 15/3; QQ='red'
        sage: QQ
        'red'
        sage: restore('QQ')
        sage: QQ
        Rational Field
        sage: x
        10
        sage: y = var('y')                                                              # needs sage.symbolic
        sage: restore('x y')
        sage: x                                                                         # needs sage.symbolic
        x
        sage: y
        Traceback (most recent call last):
        ...
        NameError: name 'y' is not defined
        sage: x = 10; y = 15/3; QQ='red'
        sage: ww = 15
        sage: restore()
        sage: x, QQ, ww                                                                 # needs sage.symbolic
        (x, Rational Field, 15)
        sage: restore('ww')
        sage: ww
        Traceback (most recent call last):
        ...
        NameError: name 'ww' is not defined
    """
    G = globals()  # this is the reason the code must be in Cython.
    if 'sage_mode' not in G:
        import sage.all
        D = sage.all.__dict__
    else:
        mode = G['sage_mode']
        if mode == 'cmdline':
            import sage.all_cmdline
            D = sage.all_cmdline.__dict__
        else:
            import sage.all
            D = sage.all.__dict__
    _restore(G, D, vars)
    import sage.calculus.calculus
    _restore(sage.calculus.calculus.syms_cur, sage.calculus.calculus.syms_default, vars)


def _restore(G, D, vars):
    if vars is None:
        for k, v in D.items():
            G[k] = v
    else:
        if isinstance(vars, str):
            if ',' in vars:
                vars = vars.split(',')
            else:
                vars = vars.split()
        for k in vars:
            if k in D:
                G[k] = D[k]
            else:
                try:
                    del G[k]      # the default value was "unset"
                except KeyError:
                    pass


def reset_interfaces():
    from sage.interfaces.quit import expect_quitall
    expect_quitall()



# sage.doctest: needs sage.combinat
"""
Ciphers
"""

#*****************************************************************************
#       Copyright (C) 2007 David Kohel <kohel@maths.usyd.edu.au>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

# Ciphers should inherit from morphisms (of sets).
# Specific cipher types will implement their functions in terms of the key

from sage.structure.element import Element


class Cipher(Element):
    """
    Cipher class
    """
    def __init__(self, parent, key):
        """
        Create a cipher.
        """
        Element.__init__(self, parent)
        self._key = key

    def __eq__(self, right):
        return type(self) is type(right) and self.parent() == right.parent() and self._key == right._key

    def _repr_(self):
        r"""
        Return the string representation of this cipher.

        EXAMPLES::

            sage: S = ShiftCryptosystem(AlphabeticStrings())
            sage: S(13)
            Shift cipher on Free alphabetic string monoid on A-Z
        """
        # return str(self._key)
        return "Cipher on %s" % self.parent().cipher_domain()

    def key(self):
        return self._key # was str(self._key)

    def domain(self):
        return self.parent().cipher_domain()

    def codomain(self):
        return self.parent().cipher_codomain()


class SymmetricKeyCipher(Cipher):
    """
    Symmetric key cipher class
    """
    def __init__(self, parent, key):
        """
        Create a symmetric cipher.
        """
        Cipher.__init__(self, parent, key)


class PublicKeyCipher(Cipher):
    """
    Public key cipher class
    """
    def __init__(self, parent, key, public=True):
        """
        Create a public key cipher.
        """
        Cipher.__init__(self, parent, key)
        self._public = public



from sage.libs.ntl.types cimport ZZX_c
from sage.rings.polynomial.polynomial_element cimport Polynomial

cdef class Polynomial_integer_dense_ntl(Polynomial):
    cdef ZZX_c _poly

    cdef Polynomial_integer_dense_ntl _new(self)



# sage.doctest: needs sage.combinat sage.modules
"""
Non-symmetric Macdonald polynomials
"""
import copy

from sage.categories.finite_enumerated_sets import FiniteEnumeratedSets
from sage.combinat.backtrack import GenericBacktracker
from sage.combinat.combinat import CombinatorialObject
from sage.combinat.combination import Combinations
from sage.combinat.permutation import Permutation
from sage.combinat.words.word import Word
from sage.misc.misc_c import prod
from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing
from sage.rings.rational_field import QQ
from sage.structure.parent import Parent
from sage.structure.unique_representation import UniqueRepresentation


class LatticeDiagram(CombinatorialObject):
    def boxes(self):
        """
        EXAMPLES::

            sage: a = LatticeDiagram([3,0,2])
            sage: a.boxes()
            [(1, 1), (1, 2), (1, 3), (3, 1), (3, 2)]
            sage: a = LatticeDiagram([2, 1, 3, 0, 0, 2])
            sage: a.boxes()
            [(1, 1), (1, 2), (2, 1), (3, 1), (3, 2), (3, 3), (6, 1), (6, 2)]
        """
        res = []
        for i in range(1, len(self) + 1):
            res += [(i, j + 1) for j in range(self[i])]
        return res

    def __getitem__(self, i):
        """
        Return the `i`-th entry of ``self``.

        Note that the indexing for lattice diagrams starts at `1`.

        EXAMPLES::

            sage: a = LatticeDiagram([3,0,2])
            sage: a[1]
            3
            sage: a[0]
            Traceback (most recent call last):
            ...
            ValueError: indexing starts at 1
            sage: a[-1]
            2
        """
        if i == 0:
            raise ValueError("indexing starts at 1")
        elif i < 0:
            i += 1
        return self._list[i - 1]

    def leg(self, i, j):
        """
        Return the leg of the box ``(i, j)`` in ``self``.

        EXAMPLES::

            sage: a = LatticeDiagram([3,1,2,4,3,0,4,2,3])
            sage: a.leg(5,2)
            [(5, 3)]
        """
        return [(i, k) for k in range(j + 1, self[i] + 1)]

    def arm_left(self, i, j):
        """
        Return the left arm of the box ``(i, j)`` in ``self``.

        EXAMPLES::

            sage: a = LatticeDiagram([3,1,2,4,3,0,4,2,3])
            sage: a.arm_left(5,2)
            [(1, 2), (3, 2)]
        """
        return [(ip, j) for ip in range(1, i) if j <= self[ip] <= self[i]]

    def arm_right(self, i, j):
        """
        Return the right arm of the box ``(i, j)`` in ``self``.

        EXAMPLES::

            sage: a = LatticeDiagram([3,1,2,4,3,0,4,2,3])
            sage: a.arm_right(5,2)
            [(8, 1)]
        """
        return [(ip, j - 1) for ip in range(i + 1, len(self) + 1)
                if j - 1 <= self[ip] < self[i]]

    def arm(self, i, j):
        """
        Return the arm of the box ``(i, j)`` in ``self``.

        EXAMPLES::

            sage: a = LatticeDiagram([3,1,2,4,3,0,4,2,3])
            sage: a.arm(5,2)
            [(1, 2), (3, 2), (8, 1)]
        """
        return self.arm_left(i, j) + self.arm_right(i, j)

    def l(self, i, j):
        """
        Return ``self[i] - j``.

        EXAMPLES::

            sage: a = LatticeDiagram([3,1,2,4,3,0,4,2,3])
            sage: a.l(5,2)
            1
        """
        return self[i] - j

    def a(self, i, j):
        """
        Return the length of the arm of the box ``(i, j)`` in ``self``.

        EXAMPLES::

            sage: a = LatticeDiagram([3,1,2,4,3,0,4,2,3])
            sage: a.a(5,2)
            3
        """
        return len(self.arm(i, j))

    def size(self):
        """
        Return the number of boxes in ``self``.

        EXAMPLES::

            sage: a = LatticeDiagram([3,1,2,4,3,0,4,2,3])
            sage: a.size()
            22
        """
        return sum(self._list)

    def flip(self):
        r"""
        Return the flip of ``self``.

        The flip map is defined as follows. Let ``r = max(self)``.
        Then ``self.flip()[i] = r - self[i]``.

        EXAMPLES::

            sage: a = LatticeDiagram([3,0,2])
            sage: a.flip()
            [0, 3, 1]
        """
        r = max(self)
        return LatticeDiagram([r - i for i in self])

    def boxes_same_and_lower_right(self, ii, jj):
        """
        Return an iterator of the boxes of ``self`` that are in row ``jj``
        but not identical with ``(ii, jj)``, or lie in the row
        ``jj - 1`` (the row directly below ``jj``; this might be the
        basement) and strictly to the right of ``(ii, jj)``.

        EXAMPLES::

            sage: a = AugmentedLatticeDiagramFilling([[1,6],[2],[3,4,2],[],[],[5,5]])
            sage: a = a.shape()
            sage: list(a.boxes_same_and_lower_right(1,1))
            [(2, 1), (3, 1), (6, 1), (2, 0), (3, 0), (4, 0), (5, 0), (6, 0)]
            sage: list(a.boxes_same_and_lower_right(1,2))
            [(3, 2), (6, 2), (2, 1), (3, 1), (6, 1)]
            sage: list(a.boxes_same_and_lower_right(3,3))
            [(6, 2)]
            sage: list(a.boxes_same_and_lower_right(2,3))
            [(3, 3), (3, 2), (6, 2)]
        """
        # Add all of the boxes in the same row
        for i in range(1, len(self) + 1):
            if self[i] >= jj and i != ii:
                yield (i, jj)

        for i in range(ii + 1, len(self) + 1):
            if self[i] >= jj - 1:
                yield (i, jj - 1)


class AugmentedLatticeDiagramFilling(CombinatorialObject):
    def __init__(self, l, pi=None):
        """
        EXAMPLES::

            sage: a = AugmentedLatticeDiagramFilling([[1,6],[2],[3,4,2],[],[],[5,5]])
            sage: a == loads(dumps(a))
            True
            sage: pi = Permutation([2,3,1]).to_permutation_group_element()
            sage: a = AugmentedLatticeDiagramFilling([[1,6],[2],[3,4,2],[],[],[5,5]],pi)
            sage: a == loads(dumps(a))
            True
        """
        if pi is None:
            pi = Permutation([1]).to_permutation_group_element()
        self._list = [[pi(i + 1)] + li for i, li in enumerate(l)]

    def __getitem__(self, i):
        """
        EXAMPLES::

            sage: a = AugmentedLatticeDiagramFilling([[1,6],[2],[3,4,2],[],[],[5,5]])
            sage: a[0]
            Traceback (most recent call last):
            ...
            ValueError: indexing starts at 1
            sage: a[1,0]
            1
            sage: a[2,0]
            2
            sage: a[3,2]
            4
            sage: a[3][2]
            4
        """
        if isinstance(i, tuple):
            i, j = i
            return self._list[i - 1][j]
        if i < 1:
            raise ValueError("indexing starts at 1")
        return self._list[i - 1]

    def shape(self):
        """
        Return the shape of ``self``.

        EXAMPLES::

            sage: a = AugmentedLatticeDiagramFilling([[1,6],[2],[3,4,2],[],[],[5,5]])
            sage: a.shape()
            [2, 1, 3, 0, 0, 2]
        """
        return LatticeDiagram([max(0, len(self[i]) - 1)
                               for i in range(1, len(self) + 1)])

    def __contains__(self, ij):
        r"""
        Return ``True`` if the box ``(i, j) (= ij)`` is in ``self``.

        Note that this does not include the basement row.

        EXAMPLES::

            sage: a = AugmentedLatticeDiagramFilling([[1,6],[2],[3,4,2],[],[],[5,5]])
            sage: (1, 1) in a
            True
            sage: (1, 0) in a
            False
        """
        i, j = ij
        return 0 < i <= len(self) and 0 < j <= len(self[i])

    def are_attacking(self, i, j, ii, jj):
        r"""
        Return ``True`` if the boxes ``(i, j)`` and ``(ii, jj)`` in ``self``
        are attacking.

        EXAMPLES::

            sage: a = AugmentedLatticeDiagramFilling([[1,6],[2],[3,4,2],[],[],[5,5]])
            sage: all( a.are_attacking(i,j,ii,jj) for (i,j),(ii,jj) in a.attacking_boxes())
            True
            sage: a.are_attacking(1,1,3,2)
            False
        """
        # If the two boxes are at the same height,
        # then they are attacking
        if j == jj:
            return True

        # Make it so that the lower box is in position i,j
        if jj < j:
            i, j, ii, jj = ii, jj, i, j

        # If the lower box is one row below and
        # strictly to the right, then they are
        # attacking.
        return j == jj - 1 and i > ii

    def boxes(self):
        """
        Return a list of the coordinates of the boxes of ``self``, including
        the 'basement row'.

        EXAMPLES::

            sage: a = AugmentedLatticeDiagramFilling([[1,6],[2],[3,4,2],[],[],[5,5]])
            sage: a.boxes()
            [(1, 1),
             (1, 2),
             (2, 1),
             (3, 1),
             (3, 2),
             (3, 3),
             (6, 1),
             (6, 2),
             (1, 0),
             (2, 0),
             (3, 0),
             (4, 0),
             (5, 0),
             (6, 0)]
        """
        return self.shape().boxes() + [(i, 0)
                                       for i in range(1, len(self.shape()) + 1)]

    def attacking_boxes(self):
        """
        Return a list of pairs of boxes in ``self`` that are attacking.

        EXAMPLES::

            sage: a = AugmentedLatticeDiagramFilling([[1,6],[2],[3,4,2],[],[],[5,5]])
            sage: a.attacking_boxes()[:5]
            [((1, 1), (2, 1)),
             ((1, 1), (3, 1)),
             ((1, 1), (6, 1)),
             ((1, 1), (2, 0)),
             ((1, 1), (3, 0))]
        """
        boxes = self.boxes()
        res = []
        for (i, j), (ii, jj) in Combinations(boxes, 2):
            if self.are_attacking(i, j, ii, jj):
                res.append(((i, j), (ii, jj)))
        return res

    def is_non_attacking(self):
        """
        Return ``True`` if ``self`` is non-attacking.

        EXAMPLES::

            sage: a = AugmentedLatticeDiagramFilling([[1,6],[2],[3,4,2],[],[],[5,5]])
            sage: a.is_non_attacking()
            True
            sage: a = AugmentedLatticeDiagramFilling([[1, 1, 1], [2, 3], [3]])
            sage: a.is_non_attacking()
            False
            sage: a = AugmentedLatticeDiagramFilling([[2,2],[1]])
            sage: a.is_non_attacking()
            False
            sage: pi = Permutation([2,1]).to_permutation_group_element()
            sage: a = AugmentedLatticeDiagramFilling([[2,2],[1]],pi)
            sage: a.is_non_attacking()
            True
        """
        return all(self[a] != self[b] for a, b in self.attacking_boxes())

    def weight(self):
        """
        Return the weight of ``self``.

        EXAMPLES::

            sage: a = AugmentedLatticeDiagramFilling([[1,6],[2],[3,4,2],[],[],[5,5]])
            sage: a.weight()
            [1, 2, 1, 1, 2, 1]
        """
        ed = self.reading_word().evaluation_dict()
        entries = list(ed)
        m = max(entries) + 1 if entries else -1
        return [ed.get(k, 0) for k in range(1, m)]

    def descents(self):
        """
        Return a list of the descents of ``self``.

        EXAMPLES::

            sage: a = AugmentedLatticeDiagramFilling([[1,6],[2],[3,4,2],[],[],[5,5]])
            sage: a.descents()
            [(1, 2), (3, 2)]
        """
        res = []
        for i, j in self.shape().boxes():
            if self[i, j] > self[i, j - 1]:
                res.append((i, j))
        return res

    def maj(self):
        """
        Return the major index of ``self``.

        EXAMPLES::

            sage: a = AugmentedLatticeDiagramFilling([[1,6],[2],[3,4,2],[],[],[5,5]])
            sage: a.maj()
            3
        """
        res = 0
        shape = self.shape()
        for i, j in self.descents():
            res += shape.l(i, j) + 1
        return res

    def reading_order(self):
        """
        Return a list of coordinates of the boxes in ``self``, starting from
        the top right, and reading from right to left.

        Note that this includes the 'basement row' of ``self``.

        EXAMPLES::

            sage: a = AugmentedLatticeDiagramFilling([[1,6],[2],[3,4,2],[],[],[5,5]])
            sage: a.reading_order()
            [(3, 3),
             (6, 2),
             (3, 2),
             (1, 2),
             (6, 1),
             (3, 1),
             (2, 1),
             (1, 1),
             (6, 0),
             (5, 0),
             (4, 0),
             (3, 0),
             (2, 0),
             (1, 0)]
        """
        boxes = self.boxes()

        def fn(ij):
            return (-ij[1], -ij[0])
        boxes.sort(key=fn)
        return boxes

    def reading_word(self):
        """
        Return the reading word of ``self``, obtained by reading the boxes
        entries of ``self`` from right to left, starting in the upper right.

        EXAMPLES::

            sage: a = AugmentedLatticeDiagramFilling([[1,6],[2],[3,4,2],[],[],[5,5]])
            sage: a.reading_word()
            word: 25465321
        """
        w = [self[i, j] for i, j in self.reading_order() if j > 0]
        return Word(w)

    def inversions(self):
        """
        Return a list of the inversions of ``self``.

        EXAMPLES::

            sage: a = AugmentedLatticeDiagramFilling([[1,6],[2],[3,4,2],[],[],[5,5]])
            sage: a.inversions()[:5]
            [((6, 2), (3, 2)),
             ((1, 2), (6, 1)),
             ((1, 2), (3, 1)),
             ((1, 2), (2, 1)),
             ((6, 1), (3, 1))]
            sage: len(a.inversions())
            25
        """
        atboxes = [set(x) for x in self.attacking_boxes()]
        res = []
        order = self.reading_order()
        for a in range(len(order)):
            i, j = order[a]
            for b in range(a + 1, len(order)):
                ii, jj = order[b]
                if self[i, j] > self[ii, jj] and set(((i, j), (ii, jj))) in atboxes:
                    res.append(((i, j), (ii, jj)))
        return res

    def _inv_aux(self):
        r"""
        EXAMPLES::

            sage: a = AugmentedLatticeDiagramFilling([[1,6],[2],[3,4,2],[],[],[5,5]])
            sage: a._inv_aux()
            7

            sage: data = [[1,6],[],[3,4,2],[5,5]]
            sage: [AugmentedLatticeDiagramFilling(data, pi=pi)._inv_aux()
            ....:  for pi in Permutations(4)]
            [4, 4, 3, 3, 2, 2, 4, 4, 2, 2, 1, 1,
             3, 3, 2, 2, 0, 0, 2, 2, 1, 1, 0, 0]
        """
        res = 0
        shape = self.shape()
        for i in range(1, len(self) + 1):
            a = self._list[i-1][0]
            res += sum(1 for j in range(i + 1, len(self) + 1)
                       if shape[i] <= shape[j] and a < self._list[j-1][0])
        return res

    def inv(self):
        r"""
        Return the inversion statistic of ``self``.

        EXAMPLES::

            sage: a = AugmentedLatticeDiagramFilling([[1,6],[2],[3,4,2],[],[],[5,5]])
            sage: a.inv()
            15
        """
        res = len(self.inversions())
        res -= sum(self.shape().a(i, j) for i, j in self.descents())
        res -= self._inv_aux()
        return res

    def coinv(self):
        """
        Return the co-inversion statistic of ``self``.

        EXAMPLES::

            sage: a = AugmentedLatticeDiagramFilling([[1,6],[2],[3,4,2],[],[],[5,5]])
            sage: a.coinv()
            2
        """
        shape = self.shape()
        return sum(shape.a(i, j) for i, j in shape.boxes()) - self.inv()

    def coeff(self, q, t):
        r"""
        Return the coefficient in front of ``self`` in the HHL formula for the
        expansion of the non-symmetric Macdonald polynomial
        ``E(self.shape())``.

        EXAMPLES::

            sage: a = AugmentedLatticeDiagramFilling([[1,6],[2],[3,4,2],[],[],[5,5]])
            sage: q,t = var('q,t')                                                      # needs sage.symbolic
            sage: a.coeff(q,t)                                                          # needs sage.symbolic
            (t - 1)^4/((q^2*t^3 - 1)^2*(q*t^2 - 1)^2)
        """
        shape = self.shape()
        return prod((1 - t) / (1 - q**(shape.l(i, j) + 1) * t**(shape.a(i, j) + 1))
                    for i, j in shape.boxes() if self[i, j] != self[i, j - 1])

    def coeff_integral(self, q, t):
        r"""
        Return the coefficient in front of ``self`` in the HHL formula for the
        expansion of the integral non-symmetric Macdonald polynomial
        ``E(self.shape())``.

        EXAMPLES::

            sage: a = AugmentedLatticeDiagramFilling([[1,6],[2],[3,4,2],[],[],[5,5]])
            sage: q,t = var('q,t')                                                      # needs sage.symbolic
            sage: a.coeff_integral(q,t)                                                 # needs sage.symbolic
            (q^2*t^3 - 1)^2*(q*t^2 - 1)^2*(t - 1)^4
        """
        res = 1
        shape = self.shape()
        for i, j in shape.boxes():
            if self[i, j] != self[i, j - 1]:
                res *= (1 - q**(shape.l(i, j) + 1) * t**(shape.a(i, j) + 1))
        for i, j in shape.boxes():
            if self[i, j] == self[i, j - 1]:
                res *= (1 - t)
        return res

    def permuted_filling(self, sigma):
        """
        Return the filling given by permuting the entries of ``self``
        by ``sigma``.

        EXAMPLES::

            sage: pi = Permutation([2,1,4,3]).to_permutation_group_element()
            sage: fill = [[2],[1,2,3],[],[3,1]]
            sage: AugmentedLatticeDiagramFilling(fill).permuted_filling(pi)
            [[2, 1], [1, 2, 1, 4], [4], [3, 4, 2]]
        """
        new_filling = []
        for col in self:
            nc = [sigma(x) for x in col[1:]]
            new_filling.append(nc)
        return AugmentedLatticeDiagramFilling(new_filling, sigma)


def NonattackingFillings(shape, pi=None):
    """
    Return the finite set of nonattacking fillings of a
    given shape.

    EXAMPLES::

        sage: NonattackingFillings([0,1,2])
        Nonattacking fillings of [0, 1, 2]
        sage: NonattackingFillings([0,1,2]).list()
        [[[1], [2, 1], [3, 2, 1]],
         [[1], [2, 1], [3, 2, 2]],
         [[1], [2, 1], [3, 2, 3]],
         [[1], [2, 1], [3, 3, 1]],
         [[1], [2, 1], [3, 3, 2]],
         [[1], [2, 1], [3, 3, 3]],
         [[1], [2, 2], [3, 1, 1]],
         [[1], [2, 2], [3, 1, 2]],
         [[1], [2, 2], [3, 1, 3]],
         [[1], [2, 2], [3, 3, 1]],
         [[1], [2, 2], [3, 3, 2]],
         [[1], [2, 2], [3, 3, 3]]]
    """
    return NonattackingFillings_shape(tuple(shape), pi)


class NonattackingFillings_shape(Parent, UniqueRepresentation):
    def __init__(self, shape, pi=None):
        """
        EXAMPLES::

            sage: n = NonattackingFillings([0,1,2])
            sage: n == loads(dumps(n))
            True
        """
        self.pi = pi
        self._shape = LatticeDiagram(shape)
        Parent.__init__(self, category=FiniteEnumeratedSets())

    def _repr_(self):
        """
        EXAMPLES::

            sage: NonattackingFillings([0,1,2])
            Nonattacking fillings of [0, 1, 2]
        """
        return "Nonattacking fillings of %s" % list(self._shape)

    def flip(self):
        """
        Return the nonattacking fillings of the flipped shape.

        EXAMPLES::

            sage: NonattackingFillings([0,1,2]).flip()
            Nonattacking fillings of [2, 1, 0]
        """
        return NonattackingFillings(list(self._shape.flip()), self.pi)

    def __iter__(self):
        """
        EXAMPLES::

            sage: NonattackingFillings([0,1,2]).list()  #indirect doctest
            [[[1], [2, 1], [3, 2, 1]],
             [[1], [2, 1], [3, 2, 2]],
             [[1], [2, 1], [3, 2, 3]],
             [[1], [2, 1], [3, 3, 1]],
             [[1], [2, 1], [3, 3, 2]],
             [[1], [2, 1], [3, 3, 3]],
             [[1], [2, 2], [3, 1, 1]],
             [[1], [2, 2], [3, 1, 2]],
             [[1], [2, 2], [3, 1, 3]],
             [[1], [2, 2], [3, 3, 1]],
             [[1], [2, 2], [3, 3, 2]],
             [[1], [2, 2], [3, 3, 3]]]
            sage: len(_)
            12

        TESTS::

            sage: NonattackingFillings([3,2,1,1]).cardinality()
            3
            sage: NonattackingFillings([3,2,1,2]).cardinality()
            4
            sage: NonattackingFillings([1,2,3]).cardinality()
            12
            sage: NonattackingFillings([2,2,2]).cardinality()
            1
            sage: NonattackingFillings([1,2,3,2]).cardinality()
            24
        """
        if sum(self._shape) == 0:
            yield AugmentedLatticeDiagramFilling([[] for _ in self._shape],
                                                 self.pi)
            return

        for z in NonattackingBacktracker(self._shape, self.pi):
            yield AugmentedLatticeDiagramFilling(z, self.pi)


class NonattackingBacktracker(GenericBacktracker):
    def __init__(self, shape, pi=None):
        """
        EXAMPLES::

            sage: from sage.combinat.sf.ns_macdonald import NonattackingBacktracker
            sage: n = NonattackingBacktracker(LatticeDiagram([0,1,2]))
            sage: n._ending_position
            (3, 2)
            sage: n._initial_state
            (2, 1)
        """
        self._shape = shape
        self._n = sum(shape)
        self._initial_data = [[None] * s for s in shape]
        if pi is None:
            pi = Permutation([1]).to_permutation_group_element()
        self.pi = pi

        # The ending position will be at the highest box
        # which is farthest right
        ending_row = max(shape)
        ending_col = len(shape) - list(reversed(list(shape))).index(ending_row)
        self._ending_position = (ending_col, ending_row)

        # Get the lowest box that is farthest left
        starting_row = 1
        nonzero = [i for i in shape if i != 0]
        starting_col = list(shape).index(nonzero[0]) + 1

        GenericBacktracker.__init__(self, self._initial_data, (starting_col, starting_row))

    def _rec(self, obj, state):
        """
        EXAMPLES::

            sage: from sage.combinat.sf.ns_macdonald import NonattackingBacktracker
            sage: n = NonattackingBacktracker(LatticeDiagram([0,1,2]))
            sage: len(list(n))
            12
            sage: obj = [ [], [None], [None, None]]
            sage: state = 2, 1
            sage: list(n._rec(obj, state))
            [([[], [1], [None, None]], (3, 1), False),
             ([[], [2], [None, None]], (3, 1), False)]
        """
        # We need to set the i,j^th entry.
        i, j = state

        # Get the next state
        new_state = self.get_next_pos(i, j)
        yld = bool(new_state is None)

        for k in range(1, len(self._shape) + 1):
            # We check to make sure that k does not
            # violate any of the attacking conditions
            if j == 1 and any(self.pi(x + 1) == k
                              for x in range(i, len(self._shape))):
                continue
            if any(obj[ii - 1][jj - 1] == k for ii, jj in
                    self._shape.boxes_same_and_lower_right(i, j) if jj != 0):
                continue

            # Fill in the in the i,j box with k+1
            obj[i - 1][j - 1] = k

            # Yield the object
            yield copy.deepcopy(obj), new_state, yld

    def get_next_pos(self, ii, jj):
        """
        EXAMPLES::

            sage: from sage.combinat.sf.ns_macdonald import NonattackingBacktracker
            sage: a = AugmentedLatticeDiagramFilling([[1,6],[2],[3,4,2],[],[],[5,5]])
            sage: n = NonattackingBacktracker(a.shape())
            sage: n.get_next_pos(1, 1)
            (2, 1)
            sage: n.get_next_pos(6, 1)
            (1, 2)
            sage: n = NonattackingBacktracker(LatticeDiagram([2,2,2]))
            sage: n.get_next_pos(3, 1)
            (1, 2)
        """
        if (ii, jj) == self._ending_position:
            return None

        for i in range(ii + 1, len(self._shape) + 1):
            if self._shape[i] >= jj:
                return i, jj

        for i in range(1, ii + 1):
            if self._shape[i] >= jj + 1:
                return i, jj + 1

        raise ValueError("we should never be here")


def _check_muqt(mu, q, t, pi=None):
    """
    EXAMPLES::

        sage: from sage.combinat.sf.ns_macdonald import _check_muqt
        sage: P, q, t, n, R, x = _check_muqt([0,0,1],None,None)
        sage: P
        Fraction Field of Multivariate Polynomial Ring in q, t over Rational Field
        sage: q
        q
        sage: t
        t
        sage: n
        Nonattacking fillings of [0, 0, 1]
        sage: R
        Multivariate Polynomial Ring in x0, x1, x2 over Fraction Field of Multivariate Polynomial Ring in q, t over Rational Field
        sage: x
        (x0, x1, x2)

    ::

        sage: q,t = var('q,t')                                                          # needs sage.symbolic
        sage: P, q, t, n, R, x = _check_muqt([0,0,1],q,None)                            # needs sage.symbolic
        Traceback (most recent call last):
        ...
        ValueError: you must specify either both q and t or neither of them

    ::

        sage: P, q, t, n, R, x = _check_muqt([0,0,1],q,2)                               # needs sage.symbolic
        Traceback (most recent call last):
        ...
        ValueError: the parents of q and t must be the same
    """
    if q is None and t is None:
        P = PolynomialRing(QQ, 'q,t').fraction_field()
        q, t = P.gens()
    elif q is not None and t is not None:
        if q.parent() != t.parent():
            raise ValueError("the parents of q and t must be the same")
        P = q.parent()
    else:
        raise ValueError("you must specify either both q and t or neither of them")
    n = NonattackingFillings(mu, pi)
    R = PolynomialRing(P, len(n._shape), 'x')
    x = R.gens()
    return P, q, t, n, R, x


def E(mu, q=None, t=None, pi=None):
    r"""
    Return the non-symmetric Macdonald polynomial in type A
    corresponding to a shape ``mu``, with basement permuted according to
    ``pi``.

    Note that if both `q` and `t` are specified, then they must have
    the same parent.

    REFERENCE:

    - J. Haglund, M. Haiman, N. Loehr.
      *A combinatorial formula for non-symmetric Macdonald polynomials*.
      :arxiv:`math/0601693v3`.

    .. SEEALSO::

        :class:`~sage.combinat.root_system.non_symmetric_macdonald_polynomials.NonSymmetricMacdonaldPolynomials`
        for a type free implementation where the polynomials are
        constructed recursively by the application of intertwining
        operators.

    EXAMPLES::

        sage: from sage.combinat.sf.ns_macdonald import E
        sage: E([0,0,0])
        1
        sage: E([1,0,0])
        x0
        sage: E([0,1,0])
        (t - 1)/(q*t^2 - 1)*x0 + x1
        sage: E([0,0,1])
        (t - 1)/(q*t - 1)*x0 + (t - 1)/(q*t - 1)*x1 + x2
        sage: E([1,1,0])
        x0*x1
        sage: E([1,0,1])
        (t - 1)/(q*t^2 - 1)*x0*x1 + x0*x2
        sage: E([0,1,1])
        (t - 1)/(q*t - 1)*x0*x1 + (t - 1)/(q*t - 1)*x0*x2 + x1*x2
        sage: E([2,0,0])
        x0^2 + (q*t - q)/(q*t - 1)*x0*x1 + (q*t - q)/(q*t - 1)*x0*x2
        sage: E([0,2,0])
        (t - 1)/(q^2*t^2 - 1)*x0^2 + (q^2*t^3 - q^2*t^2 + q*t^2 - 2*q*t + q - t + 1)/(q^3*t^3 - q^2*t^2 - q*t + 1)*x0*x1 + x1^2 + (q*t^2 - 2*q*t + q)/(q^3*t^3 - q^2*t^2 - q*t + 1)*x0*x2 + (q*t - q)/(q*t - 1)*x1*x2

        sage: [E([1,0,1], pi=pi) for pi in Permutations(3)]
        [(t - 1)/(q*t^2 - 1)*x0*x1 + x0*x2,
         x0*x1 + (q*t^2 - q*t)/(q*t^2 - 1)*x0*x2,
         (t^2 - t)/(q*t^2 - 1)*x0*x1 + x1*x2,
         x0*x1 + (q*t - q)/(q*t^2 - 1)*x1*x2,
         (t - 1)/(q*t^2 - 1)*x0*x2 + x1*x2,
         x0*x2 + (q*t^2 - q*t)/(q*t^2 - 1)*x1*x2]

    TESTS::

        sage: from sage.combinat.sf.ns_macdonald import E
        sage: E([]).parent()
        Multivariate Polynomial Ring in no variables over Fraction Field
         of Multivariate Polynomial Ring in q, t over Rational Field
    """
    P, q, t, n, R, x = _check_muqt(mu, q, t, pi)
    res = R.zero()
    for a in n:
        weight = a.weight()
        res += q**a.maj() * t**a.coinv() * a.coeff(q, t) * prod(x[i]**weight[i] for i in range(len(weight)))
    return res


def E_integral(mu, q=None, t=None, pi=None):
    r"""
    Return the integral form for the non-symmetric Macdonald
    polynomial in type A corresponding to a shape mu.

    Note that if both `q` and `t` are specified, then they must have the
    same parent.

    REFERENCE:

    - J. Haglund, M. Haiman, N. Loehr.
      *A combinatorial formula for non-symmetric Macdonald polynomials*.
      :arxiv:`math/0601693v3`.

    EXAMPLES::

        sage: from sage.combinat.sf.ns_macdonald import E_integral
        sage: E_integral([0,0,0])
        1
        sage: E_integral([1,0,0])
        (-t + 1)*x0
        sage: E_integral([0,1,0])
        (-q*t^2 + 1)*x0 + (-t + 1)*x1
        sage: E_integral([0,0,1])
        (-q*t + 1)*x0 + (-q*t + 1)*x1 + (-t + 1)*x2
        sage: E_integral([1,1,0])
        (t^2 - 2*t + 1)*x0*x1
        sage: E_integral([1,0,1])
        (q*t^3 - q*t^2 - t + 1)*x0*x1 + (t^2 - 2*t + 1)*x0*x2
        sage: E_integral([0,1,1])
        (q^2*t^3 + q*t^4 - q*t^3 - q*t^2 - q*t - t^2 + t + 1)*x0*x1 + (q*t^2 - q*t - t + 1)*x0*x2 + (t^2 - 2*t + 1)*x1*x2
        sage: E_integral([2,0,0])
        (t^2 - 2*t + 1)*x0^2 + (q^2*t^2 - q^2*t - q*t + q)*x0*x1 + (q^2*t^2 - q^2*t - q*t + q)*x0*x2
        sage: E_integral([0,2,0])
        (q^2*t^3 - q^2*t^2 - t + 1)*x0^2 + (q^4*t^3 - q^3*t^2 - q^2*t + q*t^2 - q*t + q - t + 1)*x0*x1 + (t^2 - 2*t + 1)*x1^2 + (q^4*t^3 - q^3*t^2 - q^2*t + q)*x0*x2 + (q^2*t^2 - q^2*t - q*t + q)*x1*x2

        sage: [E_integral([1,0,1], pi=pi) for pi in Permutations(3)]
        [(q*t^3 - q*t^2 - t + 1)*x0*x1 + (t^2 - 2*t + 1)*x0*x2,
         (t^2 - 2*t + 1)*x0*x1 + (q^2*t^4 - q^2*t^3 - q*t^2 + q*t)*x0*x2,
         (q*t^4 - q*t^3 - t^2 + t)*x0*x1 + (t^2 - 2*t + 1)*x1*x2,
         (t^2 - 2*t + 1)*x0*x1 + (q^2*t^3 - q^2*t^2 - q*t + q)*x1*x2,
         (q*t^3 - q*t^2 - t + 1)*x0*x2 + (t^2 - 2*t + 1)*x1*x2,
         (t^2 - 2*t + 1)*x0*x2 + (q^2*t^4 - q^2*t^3 - q*t^2 + q*t)*x1*x2]

    TESTS::

        sage: from sage.combinat.sf.ns_macdonald import E_integral
        sage: E_integral([]).parent()
        Multivariate Polynomial Ring in no variables over Fraction Field
         of Multivariate Polynomial Ring in q, t over Rational Field
    """
    P, q, t, n, R, x = _check_muqt(mu, q, t, pi)
    res = R.zero()
    for a in n:
        weight = a.weight()
        res += q**a.maj() * t**a.coinv() * a.coeff_integral(q, t) * prod(x[i]**weight[i] for i in range(len(weight)))
    return res


def Ht(mu, q=None, t=None, pi=None):
    r"""
    Return the symmetric Macdonald polynomial using the Haiman,
    Haglund, and Loehr formula.

    Note that if both `q` and `t` are specified, then they must have the
    same parent.

    REFERENCE:

    - J. Haglund, M. Haiman, N. Loehr.
      *A combinatorial formula for non-symmetric Macdonald polynomials*.
      :arxiv:`math/0601693v3`.

    EXAMPLES::

        sage: from sage.combinat.sf.ns_macdonald import Ht
        sage: HHt = SymmetricFunctions(QQ['q','t'].fraction_field()).macdonald().Ht()
        sage: Ht([0,0,1])
        x0 + x1 + x2
        sage: HHt([1]).expand(3)
        x0 + x1 + x2
        sage: Ht([0,0,2])
        x0^2 + (q + 1)*x0*x1 + x1^2 + (q + 1)*x0*x2 + (q + 1)*x1*x2 + x2^2
        sage: HHt([2]).expand(3)
        x0^2 + (q + 1)*x0*x1 + x1^2 + (q + 1)*x0*x2 + (q + 1)*x1*x2 + x2^2

    TESTS::

        sage: from sage.combinat.sf.ns_macdonald import Ht
        sage: Ht([]).parent()
        Multivariate Polynomial Ring in no variables over Fraction Field
         of Multivariate Polynomial Ring in q, t over Rational Field
    """
    P, q, t, n, R, x = _check_muqt(mu, q, t, pi)
    res = R.zero()
    for a in n:
        weight = a.weight()
        res += q**a.maj() * t**a.inv() * prod(x[i]**weight[i] for i in range(len(weight)))
    return res



"""
Freely Generated Lie Conformal Algebras

AUTHORS:

- Reimundo Heluani (2019-08-09): Initial implementation
"""

# ***************************************************************************
#       Copyright (C) 2019 Reimundo Heluani <heluani@potuz.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ***************************************************************************

from sage.algebras.lie_conformal_algebras.lie_conformal_algebra_with_basis import (
    LieConformalAlgebraWithBasis,
)
from sage.categories.cartesian_product import cartesian_product
from sage.rings.integer import Integer
from sage.sets.disjoint_union_enumerated_sets import DisjointUnionEnumeratedSets
from sage.sets.family import Family
from sage.sets.non_negative_integers import NonNegativeIntegers


class FreelyGeneratedLieConformalAlgebra(LieConformalAlgebraWithBasis):
    """
    Base class for a central extension of a freely generated Lie
    conformal algebra.

    This class provides minimal functionality, it sets up the
    family of Lie conformal algebra generators.

    .. NOTE::

        We now only accept direct sums of free modules plus
        some central generators `C_i` such that `TC_i = 0`.
    """
    def __init__(self, R, index_set=None, central_elements=None, category=None,
                 element_class=None, prefix=None, **kwds) -> None:
        """
        Initialize ``self``.

        TESTS::

            sage: V = lie_conformal_algebras.Virasoro(QQ)
            sage: TestSuite(V).run()
        """
        self._generators = Family(index_set)
        E = cartesian_product([index_set, NonNegativeIntegers()])
        if central_elements is not None:
            self._generators = DisjointUnionEnumeratedSets(
                [index_set, Family(central_elements)])
            E = DisjointUnionEnumeratedSets((cartesian_product([
                Family(central_elements), {Integer(0)}]), E))

        super().__init__(R, basis_keys=E, element_class=element_class,
                         category=category, prefix=prefix, **kwds)

        if central_elements is not None:
            self._central_elements = Family(central_elements)
        else:
            self._central_elements = ()

    def lie_conformal_algebra_generators(self):
        """
        The generators of this Lie conformal algebra.

        OUTPUT: a (possibly infinite) family of generators (as an
        `R[T]`-module) of this Lie conformal algebra.

        EXAMPLES::

            sage: Vir = lie_conformal_algebras.Virasoro(QQ)
            sage: Vir.lie_conformal_algebra_generators()
            (L, C)
            sage: V = lie_conformal_algebras.Affine(QQ,'A1')
            sage: V.lie_conformal_algebra_generators()
            (B[alpha[1]], B[alphacheck[1]], B[-alpha[1]], B['K'])
        """
        F = Family(self._generators,
                   lambda i: self.monomial((i, Integer(0))),
                   name="generator map")
        from sage.categories.sets_cat import Sets
        if F in Sets().Finite():
            return tuple(F)
        return F

    def central_elements(self) -> Family:
        """
        The central generators of this Lie conformal algebra.

        EXAMPLES::

            sage: Vir = lie_conformal_algebras.Virasoro(QQ)
            sage: Vir.central_elements()
            (C,)
            sage: V = lie_conformal_algebras.Affine(QQ, 'A1')
            sage: V.central_elements()
            (B['K'],)
        """
        return Family(self._central_elements,
                      lambda i: self.monomial((i, Integer(0))),
                      name="central_element map")



# Automatically generated by /build/sagemath/src/sage/src/sage_setup/autogen/interpreters/internal/generator.py.  Do not edit!

from cpython.ref cimport PyObject

from sage.ext.fast_callable cimport Wrapper

from sage.rings.real_mpfr cimport RealNumber
from sage.libs.mpfr cimport *
from sage.rings.complex_mpfr cimport ComplexNumber
from sage.libs.mpc cimport *

cdef class Wrapper_cc(Wrapper):
    cdef object domain
    cdef ComplexNumber domain_element
    cdef int _n_args
    cdef mpc_t* _args
    cdef int _n_constants
    cdef mpc_t* _constants
    cdef object _list_py_constants
    cdef int _n_py_constants
    cdef PyObject** _py_constants
    cdef int _n_stack
    cdef mpc_t* _stack
    cdef int _n_code
    cdef int* _code
    cdef object _domain
    cdef bint call_c(self,
                     mpc_t* args,
                     mpc_t result) except 0



"""
Root system data for type F
"""
# ****************************************************************************
#       Copyright (C) 2008-2009 Daniel Bump
#       Copyright (C) 2008-2009 Justin Walker
#       Copyright (C) 2008-2009 Nicolas M. Thiery <nthiery at users.sf.net>,
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from . import ambient_space
from sage.rings.integer_ring import ZZ
from sage.sets.family import Family

# TODO: double check that this can't be defined over ZZ


class AmbientSpace(ambient_space.AmbientSpace):
    """
    The lattice behind `F_4`.  The computations are based on Bourbaki,
    Groupes et Algèbres de Lie, Ch. 4,5,6 (planche VIII).
    """

    def __init__(self, root_system, base_ring):
        r"""
        Initialize the ambient lattice for the root system of type `F_4`.

        This essentially initializes ``Base`` with the coordinates of
        the simple roots in the canonical basis for `\RR^4`.

        EXAMPLES::

            sage: e = RootSystem(['F',4]).ambient_space()

        TESTS::

            sage: TestSuite(e).run()                                                    # needs sage.graphs
        """
        ambient_space.AmbientSpace.__init__(self, root_system, base_ring)
        v = ZZ(1)/ZZ(2)
        self.Base = [self.root(1,2,p2=1),
                     self.root(2,3,p2=1),
                     self.root(3),
                     v*(self.root(0)-self.root(1)-self.root(2)-self.root(3))]

    def dimension(self):
        """
        Return the dimension of ``self``.

        EXAMPLES::

            sage: e = RootSystem(['F',4]).ambient_space()
            sage: e.dimension()
            4
        """
        return self.root_system.cartan_type().rank()

    def root(self, i, j=None, k=None, l=None, p1=0, p2=0, p3=0, p4=0):
        """
        Compute a root from base elements of the underlying lattice.
        The arguments specify the basis elements and the signs.
        Sadly, the base elements are indexed zero-based.
        We assume that if one of the indices is not given, the rest are not as well.

        EXAMPLES::

            sage: e = RootSystem(['F',4]).ambient_space()
            sage: [ e.root(i,j,p2=1) for i in range(e.n) for j in range(i+1,e.n) ]
            [(1, -1, 0, 0), (1, 0, -1, 0), (1, 0, 0, -1), (0, 1, -1, 0), (0, 1, 0, -1), (0, 0, 1, -1)]
        """
        if i == j or j is None:
            return (-1)**p1*self.monomial(i)
        if k is None:
            return (-1)**p1*self.monomial(i) + (-1)**p2*self.monomial(j)
        if l is None:
            return (-1)**p1*self.monomial(i) + (-1)**p2*self.monomial(j)+(-1)**p3*self.monomial(k)
        return (-1)**p1*self.monomial(i) + (-1)**p2*self.monomial(j)+(-1)**p3*self.monomial(k)+(-1)**p4*self.monomial(l)

    def simple_root(self, i):
        r"""
        Return the `i`-th simple root.

        It is computed according to what Bourbaki calls the Base:

        .. MATH::

            \alpha_1 = \epsilon_2-\epsilon_3,
            \alpha_2 = \epsilon_3-\epsilon_4,
            \alpha_3 = \epsilon_4,
            \alpha_4 = \frac{1}{2} \left( \epsilon_1-\epsilon_2-\epsilon_3-\epsilon_4 \right).

        EXAMPLES::

            sage: e = RootSystem(['F',4]).ambient_space()
            sage: e.simple_roots()
            Finite family {1: (0, 1, -1, 0), 2: (0, 0, 1, -1), 3: (0, 0, 0, 1), 4: (1/2, -1/2, -1/2, -1/2)}
        """
        return self.Base[i-1]

    def negative_roots(self):
        """
        Return the negative roots.

        EXAMPLES::

            sage: e = RootSystem(['F',4]).ambient_space()
            sage: e.negative_roots()
            [(-1, 0, 0, 0),
            (0, -1, 0, 0),
            (0, 0, -1, 0),
            (0, 0, 0, -1),
            (-1, -1, 0, 0),
            (-1, 0, -1, 0),
            (-1, 0, 0, -1),
            (0, -1, -1, 0),
            (0, -1, 0, -1),
            (0, 0, -1, -1),
            (-1, 1, 0, 0),
            (-1, 0, 1, 0),
            (-1, 0, 0, 1),
            (0, -1, 1, 0),
            (0, -1, 0, 1),
            (0, 0, -1, 1),
            (-1/2, -1/2, -1/2, -1/2),
            (-1/2, -1/2, -1/2, 1/2),
            (-1/2, -1/2, 1/2, -1/2),
            (-1/2, -1/2, 1/2, 1/2),
            (-1/2, 1/2, -1/2, -1/2),
            (-1/2, 1/2, -1/2, 1/2),
            (-1/2, 1/2, 1/2, -1/2),
            (-1/2, 1/2, 1/2, 1/2)]
        """
        return [ -a for a in self.positive_roots()]

    def positive_roots(self):
        r"""
        Return the positive roots.

        These are the roots which are positive with respect to the
        lexicographic ordering of the basis elements
        (`\epsilon_1<\epsilon_2<\epsilon_3<\epsilon_4`).

        EXAMPLES::

            sage: e = RootSystem(['F',4]).ambient_space()
            sage: e.positive_roots()
            [(1, 0, 0, 0),
            (0, 1, 0, 0),
            (0, 0, 1, 0),
            (0, 0, 0, 1),
            (1, 1, 0, 0),
            (1, 0, 1, 0),
            (1, 0, 0, 1),
            (0, 1, 1, 0),
            (0, 1, 0, 1),
            (0, 0, 1, 1),
            (1, -1, 0, 0),
            (1, 0, -1, 0),
            (1, 0, 0, -1),
            (0, 1, -1, 0),
            (0, 1, 0, -1),
            (0, 0, 1, -1),
            (1/2, 1/2, 1/2, 1/2),
            (1/2, 1/2, 1/2, -1/2),
            (1/2, 1/2, -1/2, 1/2),
            (1/2, 1/2, -1/2, -1/2),
            (1/2, -1/2, 1/2, 1/2),
            (1/2, -1/2, 1/2, -1/2),
            (1/2, -1/2, -1/2, 1/2),
            (1/2, -1/2, -1/2, -1/2)]
            sage: e.rho()
            (11/2, 5/2, 3/2, 1/2)
        """
        v = ZZ(1)/ZZ(2)
        if not hasattr(self, 'PosRoots'):
            self.PosRoots = ([ self.monomial(i) for i in range(self.n) ] +
                            [ self.root(i,j,p2=0) for i in range(self.n) for j in range(i+1,self.n) ] +
                            [ self.root(i,j,p2=1) for i in range(self.n) for j in range(i+1,self.n) ] +
                            [ v*self.root(0,1,2,3,0,p2,p3,p4) for p2 in [0,1] for p3 in [0,1] for p4 in [0,1] ])
        return self.PosRoots

    def fundamental_weights(self):
        """
        Return the fundamental weights of ``self``.

        EXAMPLES::

            sage: e =  RootSystem(['F',4]).ambient_space()
            sage: e.fundamental_weights()
            Finite family {1: (1, 1, 0, 0), 2: (2, 1, 1, 0), 3: (3/2, 1/2, 1/2, 1/2), 4: (1, 0, 0, 0)}
        """
        v = ZZ(1)/ZZ(2)
        return Family({ 1: self.monomial(0)+self.monomial(1),
                        2: 2*self.monomial(0)+self.monomial(1)+self.monomial(2),
                        3: v*(3*self.monomial(0)+self.monomial(1)+self.monomial(2)+self.monomial(3)),
                        4: self.monomial(0)})


from .cartan_type import CartanType_standard_finite, CartanType_simple, CartanType_crystallographic


class CartanType(CartanType_standard_finite, CartanType_simple, CartanType_crystallographic):
    def __init__(self):
        """
        EXAMPLES::

            sage: ct = CartanType(['F',4])
            sage: ct
            ['F', 4]
            sage: ct._repr_(compact = True)
            'F4'

            sage: ct.is_irreducible()
            True
            sage: ct.is_finite()
            True
            sage: ct.is_crystallographic()
            True
            sage: ct.is_simply_laced()
            False
            sage: ct.dual()
            ['F', 4] relabelled by {1: 4, 2: 3, 3: 2, 4: 1}
            sage: ct.affine()
            ['F', 4, 1]

        TESTS::

            sage: TestSuite(ct).run()
        """
        CartanType_standard_finite.__init__(self, "F", 4)

    def _latex_(self):
        r"""
        Return a latex representation of ``self``.

        EXAMPLES::

            sage: latex(CartanType(['F',4]))
            F_4
            sage: latex(CartanType(['F',4]).dual())
            F_4 \text{ relabelled by } \left\{1 : 4, 2 : 3, 3 : 2, 4 : 1\right\}
        """
        return "F_4"

    AmbientSpace = AmbientSpace

    def coxeter_number(self):
        """
        Return the Coxeter number associated with ``self``.

        EXAMPLES::

            sage: CartanType(['F',4]).coxeter_number()
            12
        """
        return 12

    def dual_coxeter_number(self):
        """
        Return the dual Coxeter number associated with ``self``.

        EXAMPLES::

            sage: CartanType(['F',4]).dual_coxeter_number()
            9
        """
        return 9

    def dynkin_diagram(self):
        """
        Return a Dynkin diagram for type F.

        EXAMPLES::

            sage: f = CartanType(['F',4]).dynkin_diagram(); f                           # needs sage.graphs
            O---O=>=O---O
            1   2   3   4
            F4
            sage: f.edges(sort=True)                                                    # needs sage.graphs
            [(1, 2, 1), (2, 1, 1), (2, 3, 2), (3, 2, 1), (3, 4, 1), (4, 3, 1)]
        """
        from .dynkin_diagram import DynkinDiagram_class
        g = DynkinDiagram_class(self)
        for i in range(1, 4):
            g.add_edge(i, i+1)
        g.set_edge_label(2,3,2)
        return g

    def _latex_dynkin_diagram(self, label=None, node=None, node_dist=2, dual=False):
        r"""
        Return a latex representation of the Dynkin diagram.

        EXAMPLES::

            sage: print(CartanType(['F',4])._latex_dynkin_diagram())
            \draw (0 cm,0) -- (2 cm,0);
            \draw (2 cm, 0.1 cm) -- +(2 cm,0);
            \draw (2 cm, -0.1 cm) -- +(2 cm,0);
            \draw (4.0 cm,0) -- +(2 cm,0);
            \draw[shift={(3.2, 0)}, rotate=0] (135 : 0.45cm) -- (0,0) -- (-135 : 0.45cm);
            \draw[fill=white] (0 cm, 0 cm) circle (.25cm) node[below=4pt]{$1$};
            \draw[fill=white] (2 cm, 0 cm) circle (.25cm) node[below=4pt]{$2$};
            \draw[fill=white] (4 cm, 0 cm) circle (.25cm) node[below=4pt]{$3$};
            \draw[fill=white] (6 cm, 0 cm) circle (.25cm) node[below=4pt]{$4$};
            <BLANKLINE>
        """
        if label is None:
            label = lambda i: i
        if node is None:
            node = self._latex_draw_node
        ret = "\\draw (0 cm,0) -- (%s cm,0);\n" % node_dist
        ret += "\\draw (%s cm, 0.1 cm) -- +(%s cm,0);\n" % (node_dist, node_dist)
        ret += "\\draw (%s cm, -0.1 cm) -- +(%s cm,0);\n" % (node_dist, node_dist)
        ret += "\\draw (%s cm,0) -- +(%s cm,0);\n" % (node_dist*2.0, node_dist)
        if dual:
            ret += self._latex_draw_arrow_tip(1.5*node_dist-0.2, 0, 180)
        else:
            ret += self._latex_draw_arrow_tip(1.5*node_dist+0.2, 0, 0)
        for i in range(4):
            ret += node(i*node_dist, 0, label(i+1))
        return ret

    def ascii_art(self, label=None, node=None):
        """
        Return an ascii art representation of the extended Dynkin diagram.

        EXAMPLES::

            sage: print(CartanType(['F',4]).ascii_art(label = lambda x: x+2))
            O---O=>=O---O
            3   4   5   6
            sage: print(CartanType(['F',4]).ascii_art(label = lambda x: x-2))
            O---O=>=O---O
            -1  0   1   2
        """
        if label is None:
            label = lambda i: i
        if node is None:
            node = self._ascii_art_node
        ret = "{}---{}=>={}---{}\n".format(node(label(1)), node(label(2)),
                                           node(label(3)), node(label(4)))
        ret += ("{!s:4}"*4).format(label(1), label(2), label(3), label(4))
        return ret

    def dual(self):
        r"""
        Return the dual Cartan type.

        This uses that `F_4` is self-dual up to relabelling.

        EXAMPLES::

            sage: F4 = CartanType(['F',4])
            sage: F4.dual()
            ['F', 4] relabelled by {1: 4, 2: 3, 3: 2, 4: 1}

            sage: F4.dynkin_diagram()                                                   # needs sage.graphs
            O---O=>=O---O
            1   2   3   4
            F4
            sage: F4.dual().dynkin_diagram()                                            # needs sage.graphs
            O---O=>=O---O
            4   3   2   1
            F4 relabelled by {1: 4, 2: 3, 3: 2, 4: 1}
        """
        return self.relabel({1: 4, 2: 3, 3: 2, 4: 1})

    def _default_folded_cartan_type(self):
        """
        Return the default folded Cartan type.

        EXAMPLES::

            sage: CartanType(['F', 4])._default_folded_cartan_type()
            ['F', 4] as a folding of ['E', 6]
        """
        from sage.combinat.root_system.type_folded import CartanTypeFolded
        return CartanTypeFolded(self, ['E', 6], [[2], [4], [3, 5], [1, 6]])


# For unpickling backward compatibility (Sage <= 4.1)
from sage.misc.persist import register_unpickle_override
register_unpickle_override('sage.combinat.root_system.type_F',
                           'ambient_space', AmbientSpace)



"""
Homomorphisms of Lie Algebras

AUTHORS:

- Travis Scrimshaw (07-15-2013): Initial implementation
- Eero Hakavuori (08-09-2018): Morphisms defined by a generating subset
"""
# ****************************************************************************
#  Copyright (C) 2013 Travis Scrimshaw <tscrim at ucdavis.edu>
#                2018 Eero Hakavuori <eero.hakavuori at gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  https://www.gnu.org/licenses/
# *****************************************************************************

from sage.misc.cachefunc import cached_method
from sage.categories.morphism import Morphism
from sage.categories.morphism import SetMorphism
from sage.categories.homset import Homset, Hom
from sage.structure.sequence import Sequence, Sequence_generic
from sage.structure.element import get_coercion_model
from sage.structure.richcmp import richcmp
from sage.matrix.constructor import matrix
from itertools import combinations

# TODO: Refactor out common functionality with RingHomomorphism_im_gens


class LieAlgebraHomomorphism_im_gens(Morphism):
    r"""
    A homomorphism of Lie algebras.

    Let `\mathfrak{g}` and `\mathfrak{g}^{\prime}` be Lie algebras.
    A linear map `f \colon \mathfrak{g} \to \mathfrak{g}^{\prime}` is a
    homomorphism (of Lie algebras) if `f([x, y]) = [f(x), f(y)]` for all
    `x, y \in \mathfrak{g}`. Thus homomorphisms are completely determined
    by the image of the generators of `\mathfrak{g}`.

    INPUT:

    - ``parent`` -- a homset between two Lie algebras
    - ``im_gens`` -- the image of the generators of the domain
    - ``base_map`` -- a homomorphism to apply to the coefficients.
      It should be a map from the base ring of the domain to the
      base ring of the codomain.
      Note that if base_map is nontrivial then the result will
      not be a morphism in the category of Lie algebras over
      the base ring.
    - ``check`` -- whether to run checks on the validity of the defining data

    EXAMPLES::

        sage: L = LieAlgebra(QQ, 'x,y,z')
        sage: Lyn = L.Lyndon()
        sage: H = L.Hall()
        sage: phi = Lyn.coerce_map_from(H); phi
        Lie algebra morphism:
          From: Free Lie algebra generated by (x, y, z) over Rational Field in the Hall basis
          To:   Free Lie algebra generated by (x, y, z) over Rational Field in the Lyndon basis
          Defn: x |--> x
                y |--> y
                z |--> z

    You can provide a base map, creating a semilinear map that (sometimes)
    preserves the Lie bracket::

        sage: R.<x> = ZZ[]
        sage: K.<i> = NumberField(x^2 + 1)
        sage: cc = K.hom([-i])
        sage: L.<X,Y,Z,W> = LieAlgebra(K, {('X','Y'): {'Z':1}, ('X','Z'): {'W':1}})
        sage: M.<A,B,C,D> = LieAlgebra(K, {('A','B'): {'C':1}, ('A','C'): {'D':1}})
        sage: phi = L.morphism({X:A, Y:B, Z:C, W:D}, base_map=cc)
        sage: phi(X)
        A
        sage: phi(i*X)
        -i*A
        sage: all(phi(x.bracket(y)) == phi(x).bracket(phi(y)) for x,y in cartesian_product_iterator([[X,Y,Z,W],[X,Y,Z,W]]))
        True

    Note that the Lie bracket should still be preserved, even though the map is no longer linear
    over the base ring::

        sage: L.<X,Y,Z,W> = LieAlgebra(K, {('X','Y'): {'Z':i}, ('X','Z'): {'W':1}})
        sage: M.<A,B,C,D> = LieAlgebra(K, {('A','B'): {'C':-i}, ('A','C'): {'D':1}})
        sage: phi = L.morphism({X:A, Y:B, Z:C, W:D}, base_map=cc)
        sage: phi(X.bracket(Y))
        -i*C
        sage: phi(X).bracket(phi(Y))
        -i*C
    """
    def __init__(self, parent, im_gens, base_map=None, check=True):
        """
        EXAMPLES::

            sage: L = LieAlgebra(QQ, 'x,y,z')
            sage: Lyn = L.Lyndon()
            sage: H = L.Hall()
            sage: phi = Lyn.coerce_map_from(H)

        We skip the category test because the Homset's element class
        does not match this class::

            sage: TestSuite(phi).run(skip=['_test_category'])
        """
        Morphism.__init__(self, parent)
        if not isinstance(im_gens, Sequence_generic):
            if not isinstance(im_gens, (tuple, list)):
                im_gens = [im_gens]
            im_gens = Sequence(im_gens, parent.codomain(), immutable=True)
        if check:
            if len(im_gens) != len(parent.domain().lie_algebra_generators()):
                raise ValueError("number of images must equal number of generators")
            if base_map is not None and not (base_map.domain() is parent.domain().base_ring() and parent.codomain().base_ring().has_coerce_map_from(base_map.codomain())):
                raise ValueError("invalid base homomorphism")
            # TODO: Implement a (meaningful) _is_valid_homomorphism_()
            #if not parent.domain()._is_valid_homomorphism_(parent.codomain(), im_gens, base_map=base_map):
            #    raise ValueError("relations do not all (canonically) map to 0 under map determined by images of generators")
        if not im_gens.is_immutable():
            import copy
            im_gens = copy.copy(im_gens)
            im_gens.set_immutable()
        self._im_gens = im_gens
        self._base_map = base_map

    def _repr_type(self):
        """
        TESTS::

            sage: L = LieAlgebra(QQ, 'x,y,z')
            sage: Lyn = L.Lyndon()
            sage: H = L.Hall()
            sage: f = Lyn.coerce_map_from(H)
            sage: type(f)
            <class 'sage.algebras.lie_algebras.morphism.LieAlgebraHomomorphism_im_gens'>
            sage: f._repr_type()
            'Lie algebra'
        """
        return "Lie algebra"

    def im_gens(self):
        """
        Return the images of the generators of the domain.

        OUTPUT:

        - ``list`` -- a copy of the list of gens (it is safe to change this)

        EXAMPLES::

            sage: L = LieAlgebra(QQ, 'x,y,z')
            sage: Lyn = L.Lyndon()
            sage: H = L.Hall()
            sage: f = Lyn.coerce_map_from(H)
            sage: f.im_gens()
            [x, y, z]
        """
        return list(self._im_gens)

    def base_map(self):
        """
        Return the map on the base ring that is part of the defining
        data for this morphism.  May return ``None`` if a coercion is used.

        EXAMPLES::

            sage: R.<x> = ZZ[]
            sage: K.<i> = NumberField(x^2 + 1)
            sage: cc = K.hom([-i])
            sage: L.<X,Y,Z,W> = LieAlgebra(K, {('X','Y'): {'Z':1}, ('X','Z'): {'W':1}})
            sage: M.<A,B> = LieAlgebra(K, abelian=True)
            sage: phi = L.morphism({X: A, Y: B}, base_map=cc)
            sage: phi(X)
            A
            sage: phi(i*X)
            -i*A
            sage: phi.base_map()
            Ring endomorphism of Number Field in i with defining polynomial x^2 + 1
              Defn: i |--> -i
        """
        return self._base_map

    def _richcmp_(self, other, op):
        """
        Rich comparisons.

        EXAMPLES::

            sage: L = LieAlgebra(QQ, 'x,y,z')
            sage: Lyn = L.Lyndon()
            sage: H = L.Hall()
            sage: x,y,z = H.gens()
            sage: f = Hom(Lyn, H)([x,y,z])
            sage: g = Hom(Lyn, H)([x,y,z])
            sage: h = Hom(Lyn, H)([y,x,z])
            sage: f == g
            True
            sage: f is g
            False
            sage: f != h
            True
        """
        return richcmp((self._im_gens, self._base_map), (other._im_gens, other._base_map), op)

    def __hash__(self):
        """
        Return the hash of this morphism.

        EXAMPLES::

            sage: L = LieAlgebra(QQ, 'x,y,z')
            sage: Lyn = L.Lyndon()
            sage: H = L.Hall()
            sage: phi = Lyn.coerce_map_from(H)
            sage: hash(phi) == hash(phi)
            True
        """
        return hash((self._im_gens, self._base_map))

    def _repr_defn(self):
        """
        Used in constructing string representation of ``self``.

        EXAMPLES::

            sage: L = LieAlgebra(QQ, 'x,y,z')
            sage: Lyn = L.Lyndon()
            sage: H = L.Hall()
            sage: phi = Lyn.coerce_map_from(H)
            sage: print(phi._repr_defn())
            x |--> x
            y |--> y
            z |--> z
        """
        D = self.domain()
        s = '\n'.join('%s |--> %s' % (x, gen)
                      for gen, x in zip(self._im_gens, D.gens()))
        if s and self._base_map is not None:
            s += '\nwith map of base ring'
        return s

    def _call_(self, x):
        """
        Evaluate this homomorphism at ``x``.

        EXAMPLES::

            sage: L = LieAlgebra(QQ, 'x,y,z')
            sage: Lyn = L.Lyndon()
            sage: H = L.Hall()
            sage: phi = Lyn.coerce_map_from(H)
            sage: a = H.graded_basis(5)[12]; a
            [z, [y, [x, [x, z]]]]
            sage: phi(a)
            [x, [[x, z], [y, z]]] + [x, [[[x, z], z], y]]
             + [[x, y], [[x, z], z]] + [[x, [y, z]], [x, z]]
        """
        return x._im_gens_(self.codomain(), self.im_gens(), base_map=self.base_map())


class LieAlgebraHomset(Homset):
    """
    Homset between two Lie algebras.

    .. TODO::

        This is a very minimal implementation which does not
        have coercions of the morphisms.
    """
    def __init__(self, X, Y, category=None, base=None, check=True):
        """
        Initialize ``self``.

        EXAMPLES::

            sage: L = LieAlgebra(QQ, 'x,y,z')
            sage: Lyn = L.Lyndon()
            sage: H = L.Hall()
            sage: HS = Hom(Lyn, H)

        We skip the elements test since homsets are not proper parents::

            sage: TestSuite(HS).run(skip=["_test_elements"])
        """
        if base is None:
            base = X.base_ring()
        Homset.__init__(self, X, Y, category, base, check)

    def _repr_(self):
        """
        Return a string representation of ``self``.

        EXAMPLES::

            sage: L = LieAlgebra(QQ, 'x,y,z')
            sage: Lyn = L.Lyndon()
            sage: H = L.Hall()
            sage: Hom(Lyn, H)
            Set of Lie algebra morphisms
             from Free Lie algebra generated by (x, y, z) over Rational Field in the Lyndon basis
             to Free Lie algebra generated by (x, y, z) over Rational Field in the Hall basis
        """
        return "Set of Lie algebra morphisms from {} to {}".format(
                                self.domain(), self.codomain())

    def __call__(self, im_gens, check=True):
        """
        Construct a morphism from ``im_gens``.

        EXAMPLES::

            sage: L = LieAlgebra(QQ, 'x,y,z')
            sage: Lyn = L.Lyndon()
            sage: H = L.Hall()
            sage: HS = Hom(Lyn, H)
            sage: x,y,z = Lyn.gens()
            sage: phi = HS([z,x,y])
            sage: phi(x + Lyn[z,y])
            z - [x, y]

            sage: HS = Hom(Lyn, Lyn)
            sage: phi = HS([z,x,y])
            sage: a = Lyn([z, [x, [[y, z], x]]]); a
            [x, [x, [[y, z], z]]] + [x, [[x, z], [y, z]]] - [[x, [y, z]], [x, z]]
            sage: phi(a)
            [[x, [[y, z], z]], y] + 2*[[[x, z], [y, z]], y] + [[[[x, z], z], y], y]
            sage: phi(phi(phi(a))) == a
            True
        """
        if isinstance(im_gens, Morphism):
            return im_gens
        from sage.categories.lie_algebras import LieAlgebras
        if (self.domain() in LieAlgebras.FiniteDimensional.WithBasis
            and self.codomain() in LieAlgebras.FiniteDimensional.WithBasis):
            try:
                return LieAlgebraMorphism_from_generators(self.domain(), im_gens,
                                                          codomain=self.codomain(),
                                                          check=check)
            except (TypeError, ValueError):
                pass
        return LieAlgebraHomomorphism_im_gens(self, im_gens)

    @cached_method
    def zero(self):
        """
        Return the zero morphism.

        EXAMPLES::

            sage: L = LieAlgebra(QQ, 'x,y,z')
            sage: Lyn = L.Lyndon()
            sage: H = L.Hall()
            sage: HS = Hom(Lyn, H)
            sage: HS.zero()
            Generic morphism:
              From: Free Lie algebra generated by (x, y, z) over Rational Field in the Lyndon basis
              To:   Free Lie algebra generated by (x, y, z) over Rational Field in the Hall basis
        """
        return SetMorphism(self, lambda x: self.codomain().zero())

    _an_element_ = zero


class LieAlgebraMorphism_from_generators(LieAlgebraHomomorphism_im_gens):
    r"""
    A morphism between two Lie algebras defined by images of a
    generating set as a Lie algebra.

    This is the Lie algebra morphism `\phi \colon L \to K` defined on
    the chosen basis of `L` to that of `K` be using the image of some
    generating set (as a Lie algebra) of `L`.

    INPUT:

    - ``on_generators`` -- dictionary ``{X: Y}`` of the images `Y`
      in ``codomain`` of elements `X` of ``domain``
    - ``codomain`` -- a Lie algebra (optional); this is inferred
      from the values of ``on_generators`` if not given
    - ``base_map`` -- a homomorphism to apply to the coefficients.
      It should be a map from the base ring of the domain to the
      base ring of the codomain.
      Note that if base_map is nontrivial then the result will
      not be a morphism in the category of Lie algebras over
      the base ring.
    - ``check`` -- boolean (default: ``True``); if ``False`` the
      values  on the Lie brackets implied by ``on_generators`` will
      not be checked for contradictory values

    EXAMPLES:

    A reflection of one horizontal vector in the Heisenberg algebra::

        sage: L.<X,Y,Z> = LieAlgebra(QQ, {('X','Y'): {'Z':1}})
        sage: phi = L.morphism({X:-X, Y:Y}); phi
        Lie algebra endomorphism of Lie algebra on 3 generators (X, Y, Z) over Rational Field
          Defn: X |--> -X
                Y |--> Y
                Z |--> -Z

    There is no Lie algebra morphism that reflects one horizontal vector,
    but not the vertical one::

        sage: L.morphism({X:-X, Y:Y, Z:Z})
        Traceback (most recent call last):
        ...
        ValueError: this does not define a Lie algebra morphism;
         contradictory values for brackets of length 2

    Checking for mistakes can be disabled, which can produce
    invalid results::

        sage: phi = L.morphism({X:-X, Y:Y, Z:Z}, check=False); phi
        Lie algebra endomorphism of Lie algebra on 3 generators (X, Y, Z) over Rational Field
          Defn: X |--> -X
                Y |--> Y
                Z |--> Z
        sage: L[phi(X), phi(Y)] == phi(L[X,Y])
        False

    The set of keys must generate the Lie algebra::

        sage: L.morphism({X: X})
        Traceback (most recent call last):
        ...
        ValueError: [X] is not a generating set of Lie algebra on 3 generators
        (X, Y, Z) over Rational Field

    Over non-fields, generating subsets are more restricted::

        sage: L.<X,Y,Z> = LieAlgebra(ZZ, {('X','Y'): {'Z':2}})
        sage: L.morphism({X: X, Y: Y})
        Traceback (most recent call last):
        ...
        ValueError: [X, Y] is not a generating set of Lie algebra on 3
        generators (X, Y, Z) over Integer Ring

    The generators do not have to correspond to the defined generating
    set of the domain::

        sage: L.<X,Y,Z,W> = LieAlgebra(QQ, {('X','Y'): {'Z':1}, ('X','Z'): {'W':1}})
        sage: K.<A,B,C> = LieAlgebra(QQ, {('A','B'): {'C':2}})
        sage: phi = L.morphism({X+2*Y: A, X-Y: B}); phi
        Lie algebra morphism:
          From: Lie algebra on 4 generators (X, Y, Z, W) over Rational Field
          To:   Lie algebra on 3 generators (A, B, C) over Rational Field
          Defn: X |--> 1/3*A + 2/3*B
                Y |--> 1/3*A - 1/3*B
                Z |--> -2/3*C
                W |--> 0
        sage: phi(X+2*Y)
        A
        sage: phi(X)
        1/3*A + 2/3*B
        sage: phi(W)
        0
        sage: phi(Z)
        -2/3*C
        sage: all(K[phi(p), phi(q)] == phi(L[p,q])
        ....:     for p in L.basis() for q in L.basis())
        True

    A quotient type Lie algebra morphism::

        sage: K.<A,B> = LieAlgebra(SR, abelian=True)
        sage: L.morphism({X: A, Y: B})
        Lie algebra morphism:
          From: Lie algebra on 4 generators (X, Y, Z, W) over Rational Field
          To:   Abelian Lie algebra on 2 generators (A, B) over Symbolic Ring
          Defn: X |--> A
                Y |--> B
                Z |--> 0
                W |--> 0
    """
    def __init__(self, on_generators, domain=None, codomain=None, check=True, base_map=None, category=None):
        r"""
        Initialize ``self``.

        The keys of ``on_generators`` need to generate ``domain``
        as a Lie algebra.

        .. TODO::

            It might be possible to extract an explicit bracket relation that
            fails whenever some linear system fails to be solved. This would
            allow outputting an even more explicit error.

        TESTS:

        Test suite for a morphism::

            sage: L.<X,Y,Z,W> = LieAlgebra(QQ, {('X','Y'): {'Z':1}, ('X','Z'): {'W':1}})
            sage: K.<A,B,C> = LieAlgebra(QQ, {('A','B'): {'C':2}})
            sage: phi = L.morphism({X+2*Y: A, X-Y: B})
            sage: TestSuite(phi).run(skip=['_test_category'])

        Failure of inferring codomain::

            sage: L.<X> = LieAlgebra(QQ, abelian=True)
            sage: L.morphism({X: int(1)})
            Traceback (most recent call last):
            ...
            TypeError: codomain <class 'int'> is not a Lie algebra

            sage: from sage.algebras.lie_algebras.morphism import LieAlgebraMorphism_from_generators
            sage: LieAlgebraMorphism_from_generators({ZZ(1): X})
            Traceback (most recent call last):
            ...
            TypeError: domain Integer Ring is not a Lie algebra
            sage: LieAlgebraMorphism_from_generators({})
            Traceback (most recent call last):
            ...
            ValueError: no elements to infer domain from
            sage: LieAlgebraMorphism_from_generators({}, domain=L)
            Traceback (most recent call last):
            ...
            ValueError: no elements to infer codomain from

        We check that we can specify a base map to get a semi-linear morphism of Lie algebras::

            sage: R.<x> = ZZ[]
            sage: K.<i> = NumberField(x^2 + 1)
            sage: cc = K.hom([-i])
            sage: L.<X,Y,Z> = LieAlgebra(K, {('X','Y'): {'Z':i}})
            sage: M.<A,B,C> = LieAlgebra(K, {('A','B'): {'C':-i}})
            sage: phi = L.morphism({X:A, Y:B, Z:C}, base_map=cc)
            sage: phi(Z)
            C
            sage: phi(i*Z)
            -i*C
        """
        from sage.categories.lie_algebras import LieAlgebras

        cm = get_coercion_model()
        if domain is None:
            if not on_generators:
                raise ValueError("no elements to infer domain from")
            domain = cm.common_parent(*on_generators)
            if domain not in LieAlgebras:
                raise TypeError("domain %s is not a Lie algebra" % domain)
        if codomain is None:
            if not on_generators:
                raise ValueError("no elements to infer codomain from")
            codomain = cm.common_parent(*list(on_generators.values()))
            if codomain not in LieAlgebras:
                raise TypeError("codomain %s is not a Lie algebra" % codomain)

        # If the base map is nontrivial, ideally we would have machinery
        # here to determine how the base map affects the category of the
        # resulting morphism.  But for now it's not clear how to do this,
        # so we leave the category as the default for now.
        parent = Hom(domain, codomain, category=category)
        m = domain.module()
        cm = codomain.module()

        spanning_set = [X.to_vector() for X in on_generators]
        im_gens = [Y.to_vector() for Y in on_generators.values()]

        if not im_gens:
            LieAlgebraHomomorphism_im_gens.__init__(self, parent, [], base_map=base_map, check=check)
            return

        # helper function to solve linear systems Ax = b, where both x and b
        # are vectors of vectors
        def solve_linear_system(A, b, check):
            R = cm.base_ring()
            A_inv = A.solve_left(matrix.identity(A.ncols()))

            if check:
                # Verify validity of solution x = A_inv*b. Since b is a vector of
                # vectors, need to expand the matrix product by hand.
                M = A * A_inv
                for Mi, bk in zip(M.rows(), b):
                    test_bk = sum((R(Mij) * bj for Mij,bj in zip(Mi,b)), cm.zero())
                    if test_bk != bk:
                        raise ValueError("contradictory linear system")

            return [sum((R(Aij) * bk for Aij,bk in zip(Ai,b)), cm.zero())
                    for Ai in A_inv.rows()]

        bracketlength = 1
        n = 0
        while True:
            sm = m.submodule(spanning_set)
            A = matrix(sm.base_ring(), [sm.coordinate_vector(X) for X in spanning_set])
            if base_map is not None:
                A = A.apply_map(base_map)
            try:
                im_gens = solve_linear_system(A, im_gens, check)
            except ValueError:
                raise ValueError("this does not define a Lie algebra morphism; "
                                 "contradictory values for brackets of length %d"
                                 % bracketlength)

            spanning_set = list(sm.basis())
            if n == len(spanning_set):
                # no increase in dimension => no further values will be computed
                break

            # compute brackets and repeat
            bracketlength += 1
            n = len(spanning_set)
            for i,j in combinations(range(n), 2):
                # add the value of the bracket to known images
                Z = domain.bracket(spanning_set[i], spanning_set[j])
                imZ = codomain.bracket(im_gens[i], im_gens[j])
                spanning_set.append(Z.to_vector())
                im_gens.append(imZ.to_vector())

        # verify that sm is the full module m
        if not sm.has_coerce_map_from(m):
            raise ValueError("%s is not a generating set of %s"
                             % (list(on_generators), domain))

        A = matrix(m.base_ring(), spanning_set)
        im_gens = solve_linear_system(A, im_gens, check)

        LieAlgebraHomomorphism_im_gens.__init__(self, parent, im_gens, base_map=base_map, check=check)

    def _call_(self, x):
        """
        Evaluate this homomorphism at ``x``.

        EXAMPLES::

            sage: L.<X,Y,Z,W> = LieAlgebra(QQ, {('X','Y'): {'Z':1}, ('X','Z'): {'W':1}})
            sage: K.<A,B> = LieAlgebra(SR, abelian=True)
            sage: phi = L.morphism({X: A, Y: B})
            sage: phi(X)
            A
            sage: phi(Y)
            B
            sage: phi(Z)
            0
            sage: phi(W)
            0
            sage: phi(-X + 3*Y)
            -A + 3*B

            sage: K.<A,B,C> = LieAlgebra(QQ, {('A','B'): {'C':2}})
            sage: phi = L.morphism({X+2*Y: A, X-Y: B})
            sage: phi(X)
            1/3*A + 2/3*B
            sage: phi(Y)
            1/3*A - 1/3*B
            sage: phi(3*X+Y)
            4/3*A + 5/3*B
            sage: phi(3/2*W-Z+Y)
            1/3*A - 1/3*B + 2/3*C
        """
        C = self.codomain()
        bh = self._base_map
        if bh is None:
            bh = lambda t: t
        return C.sum(bh(c) * self._im_gens[i] for i, c in (x.to_vector()).items())
